"""cfn-list test runner."""
import logging
import runpy
import sys
from typing import Any, Dict  # pylint: disable=unused-import

import yaml

from ...util import argv
from .base import TestHandler

if sys.version_info.major < 3:
    from pathlib2 import Path  # pylint: disable=E
else:
    from pathlib import Path  # pylint: disable=E

TYPE_NAME = 'cfn-lint'
LOGGER = logging.getLogger('runway')


class CfnLintHandler(TestHandler):
    """Lints CFN."""

    @classmethod
    def handle(cls, name, args):
        # type: (str, Dict[str, Any]) -> None
        """Perform the actual test.

        Relies on .cfnlintrc file to be located beside the Runway config file.

        """
        cfnlintrc = Path('./.cfnlintrc')

        if not cfnlintrc.is_file():
            LOGGER.error('File must exist to use this test: %s', cfnlintrc)
            sys.exit(1)

        # prevent duplicate log messages by not passing to the root logger
        logging.getLogger('cfnlint').propagate = False
        try:
            with argv(*['cfn-lint'] + args.get('cli_args', [])):
                runpy.run_module('cfnlint', run_name='__main__')
        except SystemExit as err:  # this call will always result in SystemExit
            if err.code != 0:  # ignore zero exit codes but re-raise for non-zero
                if not (yaml.safe_load(cfnlintrc.read_text()) or
                        {}).get('templates'):
                    LOGGER.warning('cfnlintrc is missing a "templates" '
                                   'section which is required by cfn-lint')
                raise



"""Pytest fixtures and plugins."""
import pytest


@pytest.fixture
def patch_module_npm(monkeypatch):
    """Patch methods and functions used during init of RunwayModuleNpm."""
    monkeypatch.setattr('runway.module.RunwayModuleNpm.check_for_npm',
                        lambda x: None)
    monkeypatch.setattr('runway.module.warn_on_boto_env_vars', lambda x: None)



"""Runway base module."""
from __future__ import print_function

import logging
import os
import sys
from distutils.util import strtobool  # pylint: disable=E

from six.moves import input

from .base_command import BaseCommand
from .. import __version__ as version

LOGGER = logging.getLogger('runway')


class RunwayCommand(BaseCommand):
    """Base class for deployer classes."""

    @classmethod
    def path_only_contains_dirs(cls, path):
        """Return boolean on whether a path only contains directories.

        Args:
            path (str): Root path to check.

        Returns:
            bool: Wether the path only contains directories.

        """
        pathlistdir = os.listdir(path)
        if pathlistdir == []:
            return True
        if any(os.path.isfile(os.path.join(path, i)) for i in pathlistdir):
            return False
        return all(cls.path_only_contains_dirs(os.path.join(path, i))
                   for i in pathlistdir)

    @classmethod
    def get_empty_dirs(cls, path):
        """Return a list of empty directories in path.

        Args:
            path (str): Root path to check.

        Returns:
            List[str]: List of empty directory names.

        """
        empty_dirs = []
        for i in os.listdir(path):
            child_path = os.path.join(path, i)
            if i == '.git' or os.path.isfile(child_path) or \
                    os.path.islink(child_path):
                continue  # cov: ignore
            if cls.path_only_contains_dirs(child_path):
                empty_dirs.append(i)
        return empty_dirs

    @staticmethod
    def version():
        """Show current package version.

        Returns:
            str: Runway version.

        """
        print(version)

    def execute(self):
        # type: () -> None
        """Execute the command.

        Raises:
            NotImplementedError: Method not defined in subclass.

        """
        raise NotImplementedError('execute must be implimented for '
                                  'subclasses of BaseCommand.')


def get_env_from_user(default):
    """Prompt user for environment.

    Args:
        default (str): Value to return if the user would not like to provide
            their own value.

    Returns:
        str: Deploy environment.

    """
    if strtobool(input(
            'Would you like to provide a difference deploy environment? [y/n]: ')):
        response = None
        while not response:
            response = input('Deploy Environment: ')
        return response
    return default


def get_env_from_branch(branch_name, prompt_if_unexpected=False):
    """Determine environment name from git branch name.

    Args:
        branch_name (str): Git branch name to parse for the deploy environment.
        prompt_if_unexpected (bool): If the branch name is an unexpected
            format/value, the user will be prompted if they would like to
            enter a different deploy environment. (*default:* ``False``)

    Returns:
        str: Deploy environment.

    """
    if branch_name.startswith('ENV-'):
        return branch_name[4:]
    if branch_name == 'master':
        LOGGER.info('Translating git branch "master" to environment '
                    '"common"')
        return 'common'
    if prompt_if_unexpected:
        LOGGER.warning('Found unexpected branch name "%s"', branch_name)
        return get_env_from_user(branch_name)
    return branch_name


def get_env_from_directory(directory_name):
    """Determine environment name from directory name."""
    if directory_name.startswith('ENV-'):
        return directory_name[4:]
    return directory_name


def get_env(path, ignore_git_branch=False, prompt_if_unexpected=False):
    """Determine environment name.

    Args:
        path (str): Path to check for deploy environment name.
        ignore_git_branch (bool): Skip checking for git branch name.
            (*default:* ``False``)
        prompt_if_unexpected (bool): If the branch name is an unexpected
            format/value, the user will be prompted if they would like to
            enter a different deploy environment. (*default:* ``False``)

    Returns:
        str: Deploy environment.

    """
    if 'DEPLOY_ENVIRONMENT' in os.environ:
        return os.environ['DEPLOY_ENVIRONMENT']

    if ignore_git_branch:
        LOGGER.info('Skipping environment lookup from current git branch '
                    '("ignore_git_branch" is set to true in the Runway '
                    'config)')
    else:
        # These are not located with the top imports because they throw an
        # error if git isn't installed
        from git import Repo as GitRepo  # pylint: disable=import-outside-toplevel
        from git.exc import InvalidGitRepositoryError  # pylint: disable=import-outside-toplevel

        try:
            b_name = GitRepo(
                path,
                search_parent_directories=True
            ).active_branch.name
            LOGGER.info('Deriving environment name from git branch %s...',
                        b_name)
            return get_env_from_branch(b_name, prompt_if_unexpected)
        except InvalidGitRepositoryError:
            pass
        except TypeError:
            LOGGER.warning('Unable to retrieve the current git branch name!')
            LOGGER.warning('Typically this occurs when operating in a '
                           'detached-head state (e.g. what Jenkins uses when '
                           'checking out a git branch). Set the '
                           'DEPLOY_ENVIRONMENT environment variable to the '
                           'name of the logical environment (e.g. "export '
                           'DEPLOY_ENVIRONMENT=dev") to bypass this error.')
            sys.exit(1)
    LOGGER.info('Deriving environment name from directory %s...', path)
    return get_env_from_directory(os.path.basename(path))



"""Tests runway/commands/modules_command.py."""
# pylint: disable=no-self-use,protected-access,redefined-outer-name
import datetime
import logging
import os
import sys
from os import path

import boto3
import pytest
import six
import yaml
from botocore.stub import Stubber
from mock import ANY, MagicMock, call, patch

from runway.commands.modules_command import (ModulesCommand,
                                             _deployment_menu_entry,
                                             _module_menu_entry,
                                             _module_name_for_display,
                                             assume_role,
                                             load_module_opts_from_file,
                                             post_deploy_assume_role,
                                             pre_deploy_assume_role,
                                             select_modules_to_run,
                                             validate_account_alias,
                                             validate_account_credentials,
                                             validate_account_id,
                                             validate_environment)
from runway.util import MutableMap, environ

from ..factories import MockBoto3Session

MODULE = 'runway.commands.modules_command'
PATCH_RUNWAY_CONFIG = 'runway.commands.base_command.Config'


@pytest.fixture(scope='function')
def module_tag_config():
    """Return a runway.yml file for testing module tags."""
    fixture_dir = path.join(
        path.dirname(path.dirname(path.realpath(__file__))),
        'fixtures'
    )
    with open(path.join(fixture_dir, 'tag.runway.yml'), 'r') as stream:
        return yaml.safe_load(stream)


@pytest.mark.parametrize('session_name, duration, region, env_vars', [
    (None, None, 'us-east-1', None),
    ('my_session', 900, 'us-west-2', {'AWS_ACCESS_KEY_ID': 'env-var-access-key',
                                      'AWS_SECRET_ACCESS_KEY': 'env-secret-key',
                                      'AWS_SESSION_TOKEN': 'env-session-token'})
])
def test_assume_role(session_name, duration, region, env_vars, caplog, monkeypatch):
    """Test assume_role."""
    caplog.set_level(logging.INFO, logger='runway')
    role_arn = 'arn:aws:iam::123456789012:role/test'
    session = MockBoto3Session(region_name=region)
    _client, stubber = session.register_client('sts')
    monkeypatch.setattr(MODULE + '.boto3', session)

    expected = {'AWS_ACCESS_KEY_ID': 'test-aws-access-key-id',
                'AWS_SECRET_ACCESS_KEY': 'test-secret-key',
                'AWS_SESSION_TOKEN': 'test-session-token'}
    expected_sts_client = {'region_name': region}
    expected_parameters = {'RoleArn': role_arn,
                           'RoleSessionName': session_name or 'runway'}
    if duration:
        expected_parameters['DurationSeconds'] = duration
    if env_vars:
        expected_sts_client.update({k.lower(): v for k, v in env_vars.items()})

    stubber.add_response('assume_role',
                         {'Credentials': {
                             'AccessKeyId': expected['AWS_ACCESS_KEY_ID'],
                             'SecretAccessKey': expected['AWS_SECRET_ACCESS_KEY'],
                             'SessionToken': expected['AWS_SESSION_TOKEN'],
                             'Expiration': datetime.datetime.now()
                         }, 'AssumedRoleUser': {'AssumedRoleId': 'test-role-id',
                                                'Arn': role_arn + '/user'}},
                         expected_parameters)
    with stubber:
        assert assume_role(role_arn, session_name=session_name,
                           duration_seconds=duration, region=region,
                           env_vars=env_vars) == expected
    stubber.assert_no_pending_responses()
    session.assert_client_called_with('sts', **expected_sts_client)
    assert caplog.messages == ["Assuming role %s..." % role_arn]


@pytest.mark.parametrize('deployment, expected', [
    ('min_required', 'deployment_1 - sampleapp-01.cfn (us-east-1)'),
    ('min_required_multi',
     'deployment_1 - sampleapp-01.cfn, sampleapp-02.cfn (us-east-1, us-west-2)')
])
def test_deployment_menu_entry(deployment, expected, fx_deployments):
    """Test _deployment_menu_entry."""
    assert _deployment_menu_entry(fx_deployments.load(deployment)) == expected


def test_load_module_opts_from_file(tmp_path):
    """Test load_module_opts_from_file."""
    mod_opts = {'initial': 'val'}
    file_content = {'file_key': 'file_val'}
    merged = {'file_key': 'file_val', 'initial': 'val'}
    file_path = tmp_path / 'runway.module.yml'

    assert load_module_opts_from_file(str(tmp_path), mod_opts.copy()) == mod_opts

    file_path.write_text(six.u(yaml.safe_dump(file_content)))
    assert load_module_opts_from_file(str(tmp_path), mod_opts.copy()) == merged


@pytest.mark.parametrize('module, expected', [
    ('string', 'string'),
    ({'path': 'simple-dict'}, 'simple-dict'),
    ({'path': 'env-mismatch', 'environments': {'dev': True}}, 'env-mismatch'),
    ({'path': 'env-match', 'environments': {'test': 'something'}},
     'env-match (something)')
])
def test_module_menu_entry(module, expected):
    """Test _module_menu_entry."""
    assert _module_menu_entry(module, 'test') == expected


@pytest.mark.parametrize('module, expected', [
    ({'path': 'path-dict'}, 'path-dict'),
    ('string', 'string'),
    (MagicMock(path='path-attr'), 'path-attr')
])
def test_module_name_for_display(module, expected):
    """Test _module_name_for_display."""
    assert _module_name_for_display(module) == expected


def test_post_deploy_assume_role(monkeypatch, runway_context):
    """Test post_deploy_assume_role."""
    mock_restore = MagicMock()
    monkeypatch.setattr(runway_context, 'restore_existing_iam_env_vars',
                        mock_restore)

    assert not post_deploy_assume_role('test', runway_context)
    mock_restore.assert_not_called()
    assert not post_deploy_assume_role({'key': 'val'}, runway_context)
    mock_restore.assert_not_called()
    assert not post_deploy_assume_role({'post_deploy_env_revert': False},
                                       runway_context)
    mock_restore.assert_not_called()
    assert not post_deploy_assume_role({'post_deploy_env_revert': True},
                                       runway_context)
    mock_restore.assert_called_once()
    assert not post_deploy_assume_role({'post_deploy_env_revert': 'any'},
                                       runway_context)
    assert mock_restore.call_count == 2


@pytest.mark.parametrize('config', [
    ('arn:aws:iam::123456789012:role/test'),
    ({}),
    ({'post_deploy_env_revert': False,
      'arn': 'arn:aws:iam::123456789012:role/test'}),
    ({'post_deploy_env_revert': True,
      'arn': 'arn:aws:iam::123456789012:role/test',
      'duration': 900}),
    ({'post_deploy_env_revert': True,
      'session_name': 'test-custom',
      'arn': 'arn:aws:iam::123456789012:role/test',
      'duration': 900}),
    ({'test': {'arn': 'arn:aws:iam::123456789012:role/test'}}),
    ({'nope': {'arn': 'arn:aws:iam::123456789012:role/test'}}),
    ({'test': 'arn:aws:iam::123456789012:role/test'})
])
def test_pre_deploy_assume_role(config, caplog, monkeypatch, runway_context):
    """Test pre_deploy_assume_role."""
    caplog.set_level(logging.INFO, logger='runway')
    orig_creds = runway_context.boto3_credentials.copy()
    assume_response = {'AWS_ACCESS_KEY_ID': 'assumed_access_key',
                       'AWS_SECRET_ACCESS_KEY': 'assumed_secret_key',
                       'AWS_SESSION_TOKEN': 'assume_session_token'}
    expected_creds = {k.lower(): v for k, v in assume_response.items()}
    mock_save = MagicMock()
    monkeypatch.setattr(runway_context, 'save_existing_iam_env_vars', mock_save)
    runway_context.env_name = 'test'
    mock_assume = MagicMock(return_value=assume_response.copy())
    monkeypatch.setattr(MODULE + '.assume_role', mock_assume)

    assert not pre_deploy_assume_role(config, runway_context)

    if isinstance(config, dict):
        def get_value(key):
            """Get value from config."""
            env_val = config.get(runway_context.env_name)
            result = config.get(key)
            if not result and key == 'arn' and isinstance(env_val, str):
                return env_val
            if not result and isinstance(env_val, dict):
                return env_val.get(key)
            return result

        arn = get_value('arn')
        duration = get_value('duration')

        if arn:
            assert runway_context.boto3_credentials == expected_creds
            mock_assume.assert_called_once_with(role_arn=arn,
                                                session_name=config.get(
                                                    'session_name'
                                                ),
                                                duration_seconds=duration,
                                                region=runway_context.env_region,
                                                env_vars=runway_context.env_vars)
            if config.get('post_deploy_env_revert'):
                mock_save.assert_called_once()
        else:
            mock_assume.assert_not_called()
            assert runway_context.boto3_credentials == orig_creds
            assert caplog.messages == ['Skipping iam:AssumeRole; no role found'
                                       ' for environment test...']
    else:
        assert runway_context.boto3_credentials == expected_creds
        mock_assume.assert_called_once_with(role_arn=config,
                                            region=runway_context.env_region,
                                            env_vars=runway_context.env_vars)


@pytest.mark.parametrize('deployment, kwargs, mock_input, expected', [
    ('min_required', {'ci': True}, None, ['sampleapp-01.cfn']),
    ({'name': 'no_modules'}, {}, None, 1),  # config class makes this obsolete
    ('min_required', {'command': 'deploy'}, None, ['sampleapp-01.cfn']),
    ('min_required', {'command': 'destroy', 'ci': True}, None,
     ['sampleapp-01.cfn']),
    ('min_required', {'command': 'destroy'}, MagicMock(return_value='y'),
     ['sampleapp-01.cfn']),
    ('min_required', {'command': 'destroy'}, MagicMock(return_value='n'), 0),
    ('min_required_multi', {}, MagicMock(return_value='all'),
     ['sampleapp-01.cfn', 'sampleapp-02.cfn']),
    ('min_required_multi', {}, MagicMock(return_value=''), 1),
    ('min_required_multi', {}, MagicMock(return_value='0'), 1),
    ('min_required_multi', {}, MagicMock(return_value='-1'), 1),
    ('min_required_multi', {}, MagicMock(return_value='invalid'), 1),
    ('simple_parallel_module', {'command': 'destroy'},
     MagicMock(side_effect=['1', '2']), ['sampleapp-02.cfn']),
    ('tagged_multi', {'tags': ['app:test-app']}, None,
     ['sampleapp-01.cfn', 'sampleapp-02.cfn', 'sampleapp-03.cfn']),
    ('tagged_multi', {'tags': ['app:test-app', 'tier:iac'], 'ci': True},
     None, ['sampleapp-01.cfn']),
    ('tagged_multi', {'tags': ['no-match']}, None, []),
])
def test_select_modules_to_run(deployment, kwargs, mock_input, expected,
                               fx_deployments, monkeypatch, caplog):
    """Test select_modules_to_run."""
    caplog.set_level(logging.INFO, logger='runway')
    if isinstance(deployment, str):
        deployment = fx_deployments.load(deployment)
    monkeypatch.setattr(MODULE + '.input', mock_input)

    if isinstance(expected, int):
        with pytest.raises(SystemExit) as excinfo:
            assert not select_modules_to_run(deployment, **kwargs)
        assert excinfo.value.code == expected
    else:
        filtered_deployment = select_modules_to_run(deployment, **kwargs)
        result = []
        for mod in filtered_deployment.modules:
            if mod.name == 'parallel_parent':
                result.extend([m.path for m in mod.child_modules])
            else:
                result.append(mod.path)
        assert result == expected

    if mock_input:
        mock_input.assert_called()


def test_validate_account_alias(caplog):
    """Test validate_account_alias."""
    caplog.set_level(logging.INFO, logger='runway')
    alias = 'test-alias'
    iam_client = boto3.client('iam')
    stubber = Stubber(iam_client)

    stubber.add_response('list_account_aliases', {'AccountAliases': [alias]})
    stubber.add_response('list_account_aliases', {'AccountAliases': ['no-match']})

    with stubber:
        assert not validate_account_alias(iam_client, alias)
        with pytest.raises(SystemExit) as excinfo:
            assert validate_account_alias(iam_client, alias)
        assert excinfo.value.code == 1
    stubber.assert_no_pending_responses()
    assert caplog.messages == [
        'Verified current AWS account alias matches required alias {}.'.format(alias),
        'Current AWS account aliases "{}" do not match required account'
        ' alias {} in Runway config.'.format('no-match', alias)
    ]


def test_validate_account_id(caplog):
    """Test validate_account_id."""
    caplog.set_level(logging.INFO, logger='runway')
    account_id = '123456789012'
    arn = 'arn:aws:sts:us-east-1/irrelevant'
    user = 'irrelevant'
    sts_client = boto3.client('sts')
    stubber = Stubber(sts_client)

    stubber.add_response('get_caller_identity', {'UserId': user,
                                                 'Account': account_id,
                                                 'Arn': arn})
    stubber.add_response('get_caller_identity', {'UserId': user,
                                                 'Account': '012345678901',
                                                 'Arn': arn})
    stubber.add_response('get_caller_identity', {'UserId': user,
                                                 'Arn': arn})

    with stubber:
        assert not validate_account_id(sts_client, account_id)
        with pytest.raises(SystemExit) as excinfo:
            assert validate_account_id(sts_client, account_id)
        assert excinfo.value.code == 1
        with pytest.raises(SystemExit) as excinfo:
            assert validate_account_id(sts_client, account_id)
        assert excinfo.value.code == 1
    stubber.assert_no_pending_responses()
    assert caplog.messages == [
        'Verified current AWS account matches required account id %s.' % account_id,
        'Current AWS account 012345678901 does not match required account'
        ' %s in Runway config.' % account_id,
        'Error checking current account ID'
    ]


def test_validate_account_credentials(fx_deployments, monkeypatch,
                                      runway_context):
    """Test validate_account_credentials."""
    mock_alias = MagicMock()
    mock_id = MagicMock()
    account_id = '123456789012'
    alias = 'test'

    # can ignore these
    runway_context.add_stubber('iam')
    runway_context.add_stubber('sts')

    monkeypatch.setattr(MODULE + '.validate_account_alias', mock_alias)
    monkeypatch.setattr(MODULE + '.validate_account_id', mock_id)

    assert not validate_account_credentials(
        fx_deployments.load('min_required'), runway_context)
    mock_alias.assert_not_called()
    mock_id.assert_not_called()

    assert not validate_account_credentials(
        fx_deployments.load('validate_account'), runway_context)
    mock_alias.assert_called_once_with(ANY, alias)
    mock_id.assert_called_once_with(ANY, account_id)

    assert not validate_account_credentials(
        fx_deployments.load('validate_account_map'), runway_context)
    mock_alias.assert_called_with(ANY, alias)
    mock_id.assert_called_with(ANY, account_id)

    mock_alias.reset_mock()
    mock_id.reset_mock()
    runway_context.env_name = 'something-else'

    assert not validate_account_credentials(
        fx_deployments.load('validate_account_map'), runway_context)
    mock_alias.assert_not_called()
    mock_id.assert_not_called()


@pytest.mark.parametrize('env_def, strict, expected, expected_logs', [
    ({'invalid'}, False, False,
     ['test_module: skipped; unsupported type for environments "%s"' % type(set())]),
    (True, False, True, ['test_module: explicitly enabled']),
    (False, False, False, ['test_module: skipped; explicitly disabled']),
    (['123456789012/us-east-1'], False, True, []),
    (['123456789012/us-east-2'], False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ('123456789012/us-east-1', False, True, []),
    ('123456789012/us-east-2', False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ({}, False, None,
     ['test_module: environment not defined; module will determine deployment']),
    ({}, True, None,
     ['test_module: environment not defined; module will determine deployment']),
    ({'example': '111111111111/us-east-1'}, False, None,
     ['test_module: environment not in definition; module will determine deployment']),
    ({'example': '111111111111/us-east-1'}, True, False,
     ['test_module: skipped; environment not in definition']),
    ({'test': False}, False, False,
     ['test_module: skipped; explicitly disabled']),
    ({'test': True}, False, True, ['test_module: explicitly enabled']),
    ({'test': '123456789012/us-east-1'}, False, True, []),
    ({'test': '123456789012/us-east-2'}, False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ({'test': '123456789012'}, False, True, []),
    ({'test': '111111111111'}, False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ({'test': 123456789012}, False, True, []),
    ({'test': 111111111111}, False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ({'test': 'us-east-1'}, False, True, []),
    ({'test': 'us-east-2'}, False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ({'test': ['123456789012/us-east-1', '123456789012/us-east-2']}, False,
     True, []),
    ({'test': ['123456789012/us-east-2']}, False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ({'test': ['123456789012', '111111111111']}, False, True, []),
    ({'test': ['111111111111']}, False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ({'test': [123456789012, 111111111111]}, False, True, []),
    ({'test': [111111111111]}, False, False,
     ['test_module: skipped; account_id/region mismatch']),
    ({'test': ['us-east-1', 'us-east-2']}, False, True, []),
    ({'test': ['us-east-2']}, False, False,
     ['test_module: skipped; account_id/region mismatch'])
])
def test_validate_environment(env_def, strict, expected, expected_logs,
                              caplog, runway_context):
    """Test validate_environment."""
    mock_module = MutableMap(name='test_module')
    caplog.set_level(logging.DEBUG, logger='runway')
    assert validate_environment(runway_context, mock_module, env_def, strict) \
        is expected
    # all() does not give an output that can be used for troubleshooting failures
    for log in expected_logs:
        assert log in caplog.messages


@pytest.mark.usefixtures('patch_runway_config')
class TestModulesCommand(object):
    """Test runway.commands.modules_command.ModulesCommand.

    Protected methods are being patched and tested. While not best practice,
    the logic is too dense not to. It needs to be refactored into smaller
    chunks to be properly tested but, this is a "first pass" so that we
    have some coverage pre-refactor.

    """

    @patch(MODULE + '.validate_environment')
    @patch(MODULE + '.merge_nested_environment_dicts')
    @patch(MODULE + '.Path')
    @patch(MODULE + '.RunwayModuleType')
    @patch(MODULE + '.change_dir')
    @pytest.mark.parametrize('deployment, command, strict, should_validate_env', [
        ('min_required', 'deploy', True, True),
        ('min_required', 'deploy', False, False),
        ('min_required', 'destroy', False, False),
        ('min_required', 'plan', False, False),
        ('validate_account', 'deploy', True, True),
        ('validate_account', 'deploy', False, False),
        ('validate_account', 'destroy', False, False),
        ('validate_account', 'plan', False, False),
        ('simple_env_vars', 'deploy', True, True),
        ('simple_env_vars', 'deploy', False, False),
        ('simple_env_vars', 'destroy', False, False),
        ('simple_env_vars', 'plan', False, False),
        ('simple_env_vars_map', 'deploy', True, True),
        ('simple_env_vars_map', 'deploy', False, False),
        ('simple_env_vars_map', 'destroy', False, False),
        ('simple_env_vars_map', 'plan', False, False),
        ('environments_map_deprecated', 'deploy', True, True),
        ('environments_map_deprecated', 'deploy', False, False),
        ('environments_map_deprecated', 'destroy', False, False),
        ('environments_map_deprecated', 'plan', False, False),
        ('environments_map_list', 'deploy', True, True),
        ('environments_map_list', 'deploy', False, True),
        ('environments_map_list', 'destroy', False, True),
        ('environments_map_list', 'plan', False, True),
        ('environments_map_str', 'deploy', True, True),
        ('environments_map_str', 'deploy', False, True),
        ('environments_map_str', 'destroy', False, True),
        ('environments_map_str', 'plan', False, True),
    ])
    def test_deploy_module(self, mock_change_dir, mock_runwaymoduletype,
                           mock_path,
                           mock_merge_nested_environment_dicts,
                           mock_validate_environment,
                           deployment, command, strict, should_validate_env,
                           fx_deployments, monkeypatch, runway_context,
                           patch_runway_config):
        """Test _deploy_module.

        This test is not well designed but it is about as good as it can be
        given the density of logic and branches in the corresponding method.

        """
        deployment = fx_deployments.load(deployment)
        module = deployment.modules[0]
        runway_context.command = command
        mock_module_instance = MagicMock()
        patch_runway_config.future.strict_environments = strict

        mock_runwaymoduletype.return_value = mock_runwaymoduletype
        mock_runwaymoduletype.module_class.return_value = mock_module_instance

        monkeypatch.setattr(deployment, 'resolve', MagicMock())
        monkeypatch.setattr(module, 'resolve', MagicMock())
        monkeypatch.setattr(MODULE + '.load_module_opts_from_file',
                            lambda x, y: y)  # return module_opts

        assert not ModulesCommand()._deploy_module(module, deployment,
                                                   runway_context)
        # second arg should be a mock
        deployment.resolve.assert_called_once_with(runway_context, ANY)
        deployment.resolve.assert_called_once_with(runway_context, ANY)
        mock_path.assert_called_once_with(module, os.getcwd(),
                                          os.path.join(os.getcwd(),
                                          '.runway_cache'))
        if deployment.env_vars or module.env_vars:
            mock_merge_nested_environment_dicts.assert_called_once_with(
                ANY, env_name=runway_context.env_name, env_root=os.getcwd()
            )  # ANY is a dict constructed during execution
            mock_runwaymoduletype.module_class.assert_called_once_with(
                context=ANY,  # context object is altered if env_vars
                path=mock_path().module_root,
                options=ANY  # dict constructed during execution
            )
        else:
            mock_merge_nested_environment_dicts.assert_not_called()
            mock_runwaymoduletype.module_class.assert_called_once_with(
                context=runway_context,
                path=mock_path().module_root,
                options=ANY  # dict constructed during execution
            )
        if should_validate_env:
            mock_validate_environment.assert_called_once_with(
                # second arg is constructed during execution
                context=ANY,  # potentially changed runway_context
                env_def=ANY,  # merged dict of environments
                module=module,
                strict=strict
            )
        else:
            mock_validate_environment.assert_not_called()
        mock_change_dir.assert_has_calls([call(mock_path().module_root),
                                          call().__enter__(),
                                          call().__exit__(None, None, None)])
        # last two args come from module_opts
        mock_runwaymoduletype.assert_called_once_with(mock_path().module_root,
                                                      ANY, ANY)
        mock_module_instance[command].assert_called_once_with()

    @patch(MODULE + '.validate_environment')
    @patch(MODULE + '.RunwayModuleType')
    def test_deploy_module_invalid_env(self, mock_runwaymoduletype,
                                       mock_validate_environment,
                                       fx_deployments, monkeypatch,
                                       runway_context):
        """Test _deploy_module where validate_environment fails."""
        deployment = fx_deployments.load('environments_map_str')
        module = deployment.modules[0]

        monkeypatch.setattr(deployment, 'resolve', MagicMock())
        monkeypatch.setattr(module, 'resolve', MagicMock())
        monkeypatch.setattr(MODULE + '.load_module_opts_from_file',
                            lambda x, y: y)  # return module_opts
        mock_validate_environment.return_value = False

        assert not ModulesCommand()._deploy_module(module, deployment,
                                                   runway_context)
        mock_validate_environment.assert_called_once()
        mock_runwaymoduletype.assert_not_called()

    @patch(MODULE + '.hasattr')
    def test_deploy_module_missing_method(self, mock_hasattr,
                                          fx_deployments, monkeypatch,
                                          runway_context):
        """Test _deploy_module where the command method is missing."""
        deployment = fx_deployments.load('min_required')
        module = deployment.modules[0]

        mock_hasattr.return_value = False
        monkeypatch.setattr(deployment, 'resolve', MagicMock())
        monkeypatch.setattr(module, 'resolve', MagicMock())
        monkeypatch.setattr(MODULE + '.load_module_opts_from_file',
                            lambda x, y: y)  # return module_opts
        monkeypatch.setattr(MODULE + '.change_dir', MagicMock())

        with pytest.raises(SystemExit) as excinfo:
            assert ModulesCommand()._deploy_module(module, deployment,
                                                   runway_context)
        assert excinfo.value.code == 1
        # ANY is a child mock
        mock_hasattr.assert_called_once_with(ANY, runway_context.command)

    def test_execute(self):
        """Test execute."""
        with pytest.raises(NotImplementedError):
            ModulesCommand().execute()

    @patch(MODULE + '.pre_deploy_assume_role')
    @patch(MODULE + '.validate_account_credentials')
    @patch(MODULE + '.post_deploy_assume_role')
    @pytest.mark.parametrize('deployment, region, parallel', [
        ('min_required', 'us-west-2', False),
        ('min_required', 'us-west-2', True),
        ('validate_account', 'us-west-2', False),
        ('validate_account', 'us-west-2', True),
        ('simple_assume_role', 'us-west-2', False),
        ('simple_assume_role', 'us-west-2', True)
    ])
    def test_execute_deployment(self, mock_post_deploy_assume_role,
                                mock_validate_account_credentials,
                                mock_pre_deploy_assume_role,
                                deployment, region, parallel,
                                fx_deployments, monkeypatch, runway_context):
        """Test _execute_deployment."""
        # pylint: disable=no-member
        deployment = fx_deployments.load(deployment)
        monkeypatch.setattr(ModulesCommand, '_process_modules', MagicMock())

        assert not ModulesCommand()._execute_deployment(deployment,
                                                        runway_context,
                                                        region, parallel)
        ModulesCommand._process_modules.assert_called_once_with(
            deployment,
            # context will be a new instance if parallel
            runway_context if not parallel else ANY
        )

        if parallel:
            assert runway_context.env_region == 'us-east-1'
        else:
            assert runway_context.env_region == region
            assert runway_context.env_vars['AWS_DEFAULT_REGION'] == region
            assert runway_context.env_vars['AWS_REGION'] == region

        if deployment.assume_role:
            mock_pre_deploy_assume_role.assume_called_once_with(
                deployment.assume_role,
                # context will be a new instance if parallel
                runway_context if not parallel else ANY
            )
            mock_post_deploy_assume_role.assume_called_once_with(
                deployment.assume_role,
                # context will be a new instance if parallel
                runway_context if not parallel else ANY
            )
        if deployment.account_id or deployment.account_alias:
            mock_validate_account_credentials.assert_called_once_with(
                deployment,
                # context will be a new instance if parallel
                runway_context if not parallel else ANY
            )

    @patch(MODULE + '.merge_dicts')
    @patch(MODULE + '.merge_nested_environment_dicts')
    def test_process_deployments(self, mock_merge_nested_environment_dicts,
                                 mock_merge_dicts, fx_config, monkeypatch,
                                 runway_context):
        """Test _process_deployments."""
        config = fx_config.load('min_required')  # single deployment for tests
        deployment = config.deployments[0]
        monkeypatch.setattr(deployment, 'resolve', MagicMock())
        monkeypatch.setattr(ModulesCommand, 'runway_config', config)
        monkeypatch.setattr(ModulesCommand, '_execute_deployment', MagicMock())

        ModulesCommand()._process_deployments(config.deployments,
                                              runway_context)

        deployment.resolve.assert_called_once_with(runway_context,
                                                   config.variables,
                                                   pre_process=True)
        # pylint: disable=no-member
        ModulesCommand._execute_deployment.assert_has_calls([
            call(deployment, runway_context, region)
            for region in deployment.regions
        ])
        mock_merge_nested_environment_dicts.assert_not_called()
        mock_merge_dicts.assert_not_called()

        copy_env_vars = runway_context.env_vars.copy()
        deployment._env_vars = MagicMock()
        deployment._env_vars.value = {'key': 'val'}
        mock_merge_nested_environment_dicts.return_value = {'state': 'env'}
        mock_merge_dicts.return_value = {'state': 'merged'}
        ModulesCommand()._process_deployments(config.deployments,
                                              runway_context)
        mock_merge_nested_environment_dicts.assert_called_once_with(
            {'key': 'val'},
            env_name=runway_context.env_name,
            env_root=os.getcwd()
        )
        mock_merge_dicts.assert_called_once_with(copy_env_vars,
                                                 {'state': 'env'})
        assert runway_context.env_vars == {'state': 'merged'}
        assert ModulesCommand._execute_deployment.call_count == 2

        ModulesCommand._execute_deployment.reset_mock()
        deployment.modules = []
        ModulesCommand({'--tag': ['something']})._process_deployments(
            config.deployments,
            runway_context
        )
        ModulesCommand._execute_deployment.assert_not_called()

        config = fx_config.load('min_required')
        deployment = config.deployments[0]
        monkeypatch.setattr(deployment, 'resolve', MagicMock())
        deployment._regions = MagicMock()
        deployment._regions.value = []
        with pytest.raises(SystemExit) as excinfo:
            ModulesCommand()._process_deployments(config.deployments,
                                                  runway_context)
        assert excinfo.value.code == 1

    @patch('runway.commands.modules_command.concurrent.futures')
    @pytest.mark.skipif(sys.version_info.major < 3,
                        reason='only supported by python 3')
    @pytest.mark.parametrize('config, use_concurrent', [
        ('simple_parallel_regions.1', True),
        ('simple_parallel_regions.1', False),
        ('simple_parallel_regions.2', True),
        ('simple_parallel_regions.2', False)
    ])
    def test_process_deployments_parallel(self, mock_futures,
                                          config, use_concurrent,
                                          fx_config, monkeypatch,
                                          runway_context):
        """Test _process_deployments with parallel regions."""
        # pylint: disable=no-member
        config = fx_config.load(config)  # single deployment for tests
        deployment = config.deployments[0]
        executor = MagicMock()
        mock_futures.ProcessPoolExecutor.return_value = executor
        runway_context.use_concurrent = use_concurrent
        monkeypatch.setattr(deployment, 'resolve', MagicMock())
        monkeypatch.setattr(ModulesCommand, 'runway_config', config)
        monkeypatch.setattr(ModulesCommand, '_execute_deployment', MagicMock())

        ModulesCommand()._process_deployments(config.deployments,
                                              runway_context)

        deployment.resolve.assert_called_once_with(runway_context,
                                                   config.variables,
                                                   pre_process=True)

        if not use_concurrent:
            ModulesCommand._execute_deployment.assert_called()
            return

        mock_futures.ProcessPoolExecutor.assert_called_once_with(
            max_workers=runway_context.max_concurrent_regions
        )
        executor.submit.has_calls([
            (ModulesCommand._execute_deployment, deployment, runway_context,
             region, True) for region in deployment.parallel_regions
        ])
        mock_futures.wait.called_once()
        assert executor.submit.return_value.result.call_count == \
            len(deployment.parallel_regions)

    @pytest.mark.parametrize('deployment', [('min_required'),
                                            ('min_required_multi')])
    def test_process_modules(self, deployment, fx_deployments, monkeypatch,
                             runway_context):
        """Test _process_modules."""
        # pylint: disable=no-member
        deployment = fx_deployments.load(deployment)
        monkeypatch.setattr(ModulesCommand, '_deploy_module', MagicMock())

        assert not ModulesCommand()._process_modules(deployment, runway_context)
        ModulesCommand._deploy_module.assert_has_calls([
            call(module, deployment, runway_context)
            for module in deployment.modules
        ])

    @patch('runway.commands.modules_command.concurrent.futures')
    @pytest.mark.skipif(sys.version_info.major < 3,
                        reason='only supported by python 3')
    @pytest.mark.parametrize('deployment, use_concurrent', [
        ('simple_parallel_module', True),
        ('simple_parallel_module', False),
    ])
    def test_process_modules_parallel(self, mock_futures, deployment,
                                      use_concurrent, fx_deployments,
                                      monkeypatch, runway_context):
        """Test _process_modules with parallel regions."""
        # pylint: disable=no-member
        deployment = fx_deployments.load(deployment)
        executor = MagicMock()
        mock_futures.ProcessPoolExecutor.return_value = executor
        runway_context.use_concurrent = use_concurrent

        monkeypatch.setattr(ModulesCommand, '_deploy_module', MagicMock())

        assert not ModulesCommand()._process_modules(deployment, runway_context)
        if use_concurrent:
            mock_futures.ProcessPoolExecutor.assert_called_once_with(
                max_workers=runway_context.max_concurrent_modules
            )
            executor.submit.assert_has_calls([
                call(ModulesCommand._deploy_module, x, deployment,
                     runway_context)
                for x in deployment.modules[0].child_modules
            ])
            mock_futures.wait.assert_called_once()
            assert executor.submit.return_value.result.call_count == \
                len(deployment.modules[0].child_modules)
            ModulesCommand._deploy_module.assert_called_once_with(
                deployment.modules[1],
                deployment,
                runway_context
            )
        else:
            ModulesCommand._deploy_module.assert_has_calls([
                call(module, deployment, runway_context)
                for module in deployment.modules[0].child_modules
            ] + [call(deployment.modules[1], deployment, runway_context)])

    def test_reverse_deployments(self, fx_deployments, monkeypatch):
        """Test reverse_deployments."""
        deployment = fx_deployments.load('min_required')
        monkeypatch.setattr(deployment, 'reverse', MagicMock())
        mock_deployments = MagicMock()
        mock_deployments.__iter__.return_value = iter([deployment])

        assert ModulesCommand.reverse_deployments(None) == []
        assert ModulesCommand.reverse_deployments(mock_deployments)
        mock_deployments.reverse.assert_called_once()
        deployment.reverse.assert_called_once()

    @patch(MODULE + '.select_modules_to_run')
    @patch(MODULE + '.get_env')
    @patch(MODULE + '.Context')
    @pytest.mark.parametrize('config, command, cli_args, mock_input', [
        ('min_required', 'deploy', {}, None),
        ('min_required', 'destroy', {}, None),
        ('min_required', 'plan', {}, None),
        ('min_required', 'deploy', {'--tag': ['something']}, None),
        ('min_required', 'destroy', {'--tag': ['something']}, None),
        ('min_required', 'plan', {'--tag': ['something']}, None),
        ('min_required', 'deploy', {}, MagicMock(return_value='y')),
        ('min_required', 'destroy', {}, MagicMock(return_value='y')),
        ('min_required', 'plan', {}, MagicMock(return_value='y')),
        ('min_required', 'deploy', {'--tag': ['something']},
         MagicMock(return_value='y')),
        ('min_required', 'destroy', {'--tag': ['something']},
         MagicMock(return_value='y')),
        ('min_required', 'plan', {'--tag': ['something']},
         MagicMock(return_value='y')),
        ('min_required', 'deploy', {}, MagicMock(return_value='n')),
        ('min_required', 'destroy', {}, MagicMock(return_value='n')),
        ('min_required', 'plan', {}, MagicMock(return_value='n')),
        ('min_required', 'deploy', {'--tag': ['something']},
         MagicMock(return_value='n')),
        ('min_required', 'destroy', {'--tag': ['something']},
         MagicMock(return_value='n')),
        ('min_required', 'plan', {'--tag': ['something']},
         MagicMock(return_value='n'))
    ])
    def test_run(self, mock_context, mock_get_env, mock_select_modules_to_run,
                 config, command, cli_args, mock_input,
                 fx_config, monkeypatch):
        """Test run."""
        config = fx_config.load(config)
        config.future.strict_environments = True

        mock_get_env.return_value = 'test'
        mock_context.return_value = mock_context
        mock_context.is_interactive = bool(mock_input)
        mock_context.is_noninteractive = not bool(mock_input)  # True if None
        mock_context.env_vars = {}

        monkeypatch.setattr(MODULE + '.input', mock_input)
        monkeypatch.setattr(ModulesCommand, 'reverse_deployments',
                            MagicMock(return_value=['reversed-deployments']))
        monkeypatch.setattr(ModulesCommand, 'runway_config', config)
        monkeypatch.setattr(ModulesCommand, 'select_deployment_to_run',
                            MagicMock(return_value=['deployment']))
        monkeypatch.setattr(ModulesCommand, '_process_deployments', MagicMock())

        if command == 'destroy' and mock_input and mock_input.return_value != 'y':
            with pytest.raises(SystemExit) as excinfo:
                ModulesCommand(cli_args).run(command=command)
            mock_input.assert_called_once_with('Proceed?: ')
            assert excinfo.value.code == 0
            return  # end here since a lot of the rest will be skipped

        env_vars = {}
        if not mock_input:
            env_vars['CI'] = '1'

        with environ(env_vars):
            ModulesCommand(cli_args).run(command=command)

            mock_get_env.assert_called_once_with(
                os.getcwd(), config.ignore_git_branch,
                prompt_if_unexpected=bool(mock_input)
            )
            mock_context.assert_called_once_with(env_name='test',
                                                 env_region=None,
                                                 env_root=os.getcwd(),
                                                 env_vars=os.environ,
                                                 command=command)

        assert mock_context.env_vars['RUNWAYCONFIG'] == './runway.yml'

        if command == 'destroy' and mock_input:
            mock_input.assert_called_once_with('Proceed?: ')
            ModulesCommand.reverse_deployments.assert_called_once()

        if mock_context.is_noninteractive or cli_args.get('--tag'):
            mock_select_modules_to_run.assert_called()
        else:
            ModulesCommand.select_deployment_to_run.assert_called_once_with(
                config.deployments, command)
            mock_select_modules_to_run.assert_called_once_with(
                'deployment', cli_args.get('--tag'), command,
                mock_context.is_noninteractive, mock_context.env_name
            )
        # pylint: disable=no-member
        ModulesCommand._process_deployments.assert_called_once_with(ANY,
                                                                    mock_context)

    @pytest.mark.parametrize('config, command, mock_input, expected', [
        ('min_required', 'build', None, ['deployment_1']),
        ('min_required', 'destroy', None, ['deployment_1']),
        ('min_required_multi', 'deploy', MagicMock(return_value='1'),
         ['deployment_1']),
        ('min_required_multi', 'destroy', MagicMock(return_value='1'),
         ['deployment_1']),
        ('min_required_multi', 'deploy', MagicMock(return_value='2'),
         ['deployment_2']),
        ('min_required_multi', 'destroy', MagicMock(return_value='2'),
         ['deployment_2']),
        ('min_required_multi', 'deploy', MagicMock(return_value='all'),
         ['deployment_1', 'deployment_2']),
        ('min_required_multi', 'destroy', MagicMock(return_value='all'),
         ['deployment_1', 'deployment_2']),
        ('min_required_multi', 'deploy', MagicMock(return_value=''), 1),
        ('min_required_multi', 'destroy', MagicMock(return_value=''), 1),
    ])
    def test_select_deployment_to_run(self, config, command, mock_input, expected,
                                      fx_config, monkeypatch):
        """Test select_deployment_to_run."""
        config = fx_config.load(config)
        monkeypatch.setattr(MODULE + '.input', mock_input)

        assert ModulesCommand.select_deployment_to_run([], command) == []

        if isinstance(expected, int):
            with pytest.raises(SystemExit) as excinfo:
                assert not ModulesCommand.select_deployment_to_run(
                    config.deployments,
                    command
                )
            assert excinfo.value.code == expected
        else:
            result = ModulesCommand.select_deployment_to_run(
                config.deployments,
                command
            )
            assert [r.name for r in result] == expected
        if mock_input:
            mock_input.assert_called_once_with(
                'Enter number of deployment to run (or "all"): '
            )



"""Test runway.commands.base_command."""
# pylint: disable=no-self-use,protected-access
import os

import pytest
from mock import call

from runway.commands.base_command import BaseCommand
from runway.util import MutableMap

MODULE = 'runway.commands.base_command'
PATCH_RUNWAY_CONFIG = MODULE + '.Config'


class TestBaseCommand(object):
    """Test runway.commands.base_command.BaseCommand."""

    def test_class_attributes(self):
        """Test class attributes."""
        assert BaseCommand.SKIP_FIND_CONFIG is False
        # wording of the message does not have to be exact
        assert 'deprecated' in BaseCommand.DEPRECATION_MSG
        assert 'next major release' in BaseCommand.DEPRECATION_MSG

    def test_execute(self, patch_runway_config):
        """Test execute."""
        patch_runway_config.find_config_file.return_value = 'runway.yml'
        with pytest.raises(NotImplementedError) as excinfo:
            assert BaseCommand(env_root='./',
                               runway_config_dir='./').execute()
        assert str(excinfo.value) == \
            'execute must be implimented for subclasses of BaseCommand.'

    def test_init(self, patch_runway_config):
        """Test init and the attributes it sets."""
        cli_args = {
            '--tag': ['something']
        }
        runway_yml = os.path.join(os.getcwd(), 'runway.yml')
        patch_runway_config.find_config_file.return_value = runway_yml
        obj = BaseCommand(cli_args)

        assert obj._cli_arguments == cli_args
        assert obj._runway_config is None
        assert obj.env_root == os.getcwd()
        assert obj.runway_config_path == runway_yml
        patch_runway_config.find_config_file \
            .assert_called_once_with(os.getcwd())

    def test_init_args_optional(self, patch_runway_config, tmp_path):
        """Test init optional args."""
        runway_config_dir = tmp_path / 'some_dir'
        runway_yml = str(runway_config_dir / 'runway.yml')
        patch_runway_config.find_config_file.return_value = runway_yml
        obj = BaseCommand(env_root=str(tmp_path),
                          runway_config_dir=str(runway_config_dir))

        assert obj._cli_arguments == {}
        assert obj._runway_config is None
        assert obj.env_root == str(tmp_path)
        assert obj.runway_config_path == runway_yml
        patch_runway_config.find_config_file \
            .assert_called_once_with(str(runway_config_dir))

    def test_init_exception_systemexit(self, patch_runway_config):
        """Test init SystemExit raised by Config."""
        cwd = os.getcwd()
        cwd_parent = os.path.dirname(cwd)
        env_root = os.path.join(cwd, 'tests')
        patch_runway_config.find_config_file.side_effect = [SystemExit,
                                                            'runway.yml']
        obj = BaseCommand(env_root=env_root)

        assert obj.runway_config_path == 'runway.yml'
        assert obj.env_root == cwd_parent
        patch_runway_config.find_config_file.assert_has_calls([call(env_root),
                                                               call(cwd_parent)])

    def test_init_skip_find_config(self, monkeypatch, patch_runway_config):
        """Test init where SKIP_FIND_CONFIG=True (subclasses)."""
        monkeypatch.setattr(BaseCommand, 'SKIP_FIND_CONFIG', True)
        obj = BaseCommand()

        assert obj._runway_config is None
        assert obj.runway_config_path is None
        patch_runway_config.find_config_file.assert_not_called()

    def test_runway_config(self, patch_runway_config):
        """Test runway config."""
        obj = BaseCommand()
        assert obj.runway_config == patch_runway_config
        # ensure value is properly cached
        assert obj._runway_config == obj.runway_config
        patch_runway_config.load_from_file \
            .assert_called_once_with('./runway.yml')

    def test_runway_vars(self, patch_runway_config):
        """Test runway_vars."""
        obj = BaseCommand()
        assert isinstance(obj.runway_config, MutableMap)
        assert obj.runway_vars.data == {}
        patch_runway_config.load_from_file.assert_called_once()



"""Test classes."""
import io
import os
import sys

import boto3
import yaml
from botocore.stub import Stubber
from mock import MagicMock
from six import string_types

from runway.cfngin.context import Context as CFNginContext
from runway.config import DeploymentDefinition
from runway.context import Context as RunwayContext
from runway.util import MutableMap


class MockBoto3Session(object):
    """Mock class that acts like a boto3.session.

    Must be preloaded with stubbers.

    """

    def __init__(self,
                 clients=None,
                 aws_access_key_id=None,
                 aws_secret_access_key=None,
                 aws_session_token=None,
                 profile_name=None,
                 region_name=None):
        """Instantiate class.

        Args:
            clients (Dict[str, Any]): Clients that have already been stubbed.
            aws_access_key_id (Optional[str]): Same as boto3.Session.
            aws_secret_access_key (Optional[str]): Same as boto3.Session.
            aws_session_token (Optional[str]): Same as boto3.Session.
            profile_name (Optional[str]): Same as boto3.Session.
            region_name (Optional[str]): Same as boto3.Session.

        """
        self._clients = clients or {}
        self._client_calls = {}
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.profile_name = profile_name
        self.region_name = region_name

    def assert_client_called_with(self, service_name, **kwargs):
        """Assert a client was created with the provided kwargs."""
        key = '{}.{}'.format(service_name, kwargs.get('region_name',
                                                      self.region_name))
        assert self._client_calls[key] == kwargs

    def client(self, service_name, **kwargs):
        """Return a stubbed client.

        Args:
            service_name (str): The name of a service, e.g. 's3' or 'ec2'.

        Returns:
            Stubbed boto3 client.

        Raises:
            KeyError: Client was not stubbed from Context before trying to use.

        """
        key = '{}.{}'.format(service_name, kwargs.get('region_name',
                                                      self.region_name))
        self._client_calls[key] = kwargs
        return self._clients[key]

    def register_client(self, service_name, region_name=None):
        """Register a client for the boto3 session.

        Args:
            service_name (str): The name of a service, e.g. 's3' or 'ec2'.
            region_name (Optional[str]): AWS region.

        """
        key = '{}.{}'.format(service_name, region_name or self.region_name)
        client = boto3.client(service_name,
                              region_name=region_name or self.region_name)
        stubber = Stubber(client)
        self._clients[key] = client
        return client, stubber

    def service(self, service_name, region_name=None):
        """Not implimented."""
        raise NotImplementedError


class MockCFNginContext(CFNginContext):
    """Subclass CFNgin context object for tests."""

    def __init__(self,
                 environment=None,
                 boto3_credentials=None,
                 stack_names=None,
                 config=None,
                 config_path=None,
                 region='us-east-1',
                 force_stacks=None):
        """Instantiate class."""
        if not boto3_credentials:
            boto3_credentials = {}

        self._boto3_test_client = MutableMap()
        self._boto3_test_stubber = MutableMap()

        # used during init process
        self.__boto3_credentials = boto3_credentials
        self.s3_stubber = self.add_stubber('s3', region=region)

        super(MockCFNginContext, self).__init__(environment=environment,
                                                boto3_credentials=boto3_credentials,
                                                stack_names=stack_names,
                                                config=config,
                                                config_path=config_path,
                                                region=region,
                                                force_stacks=force_stacks)

    def add_stubber(self, service_name, region=None):
        """Add a stubber to context.

        Args:
            service_name (str): The name of a service, e.g. 's3' or 'ec2'.

        """
        key = '{}.{}'.format(service_name, region or self.region)

        self._boto3_test_client[key] = boto3.client(
            service_name,
            region_name=region or self.region,
            **self.__boto3_credentials
        )
        self._boto3_test_stubber[key] = Stubber(
            self._boto3_test_client[key]
        )
        return self._boto3_test_stubber[key]

    def get_session(self, profile=None, region=None):
        """Wrap get_session to enable stubbing."""
        return MockBoto3Session(clients=self._boto3_test_client,
                                profile_name=profile,
                                region_name=region or self.region)


# TODO use pytest-subprocess for when dropping python 2
class MockProcess(object):  # pylint: disable=too-few-public-methods
    """Instances of this class are the return_value of patched subprocess.Popen."""

    def __init__(self, returncode=0, stdout=None, universal_newlines=True):
        """Instantiate class.

        Args:
            returncode (int): Code that will be returned when the process exits.
            stdout (Optional[Union[bytes, str, List[str], Tuple(str, ...)]]):
                Content to be written accessable as stdout on the process.
            universal_newlines (bool): Use universal line endings.

        """
        self.returncode = returncode
        self.text_mode = bool(universal_newlines)
        self.stdout = self._prepare_buffer(stdout)

        self._wait = MagicMock(return_value=self.returncode)

    @property
    def wait(self):
        """Mock wait method as a property to call the stored MagicMock."""
        return self._wait

    def _prepare_buffer(self, data):
        """Prepare buffer for stdout/stderr.

        Args:
            data (Optional[Union[bytes, str, List[str], Tuple(str, ...)]]):
                Content to be written accessable as stdout on the process.

        Returns:
            Union[BytesIO, StringIO]

        """
        result = None
        linesep = os.linesep

        if isinstance(data, (list, tuple)):
            result = linesep.join(data)

        if isinstance(data, string_types):
            result = data

        if result:
            if not result.endswith(linesep):
                result += linesep
            if self.text_mode:
                result.replace('\r\n', '\n')

        io_base = io.StringIO() if self.text_mode else io.BytesIO()

        if result:
            if sys.version_info.major < 3:
                io_base.write(result.decode('UTF-8'))
            else:
                io_base.write(result)
            io_base.seek(0)  # return to the begining of the stream
        return io_base


class MockRunwayConfig(MutableMap):
    """Mock Runway config object."""

    def __init__(self, **kwargs):
        """Instantiate class."""
        super(MockRunwayConfig, self).__init__()
        self._kwargs = kwargs
        self.deployments = []
        self.future = MagicMock()
        self.tests = []
        self.ignore_git_branch = False
        self.variables = MutableMap()

        # classmethods
        self.find_config_file = MagicMock(name='find_config_file',
                                          return_value='./runway.yml')
        self.load_from_file = MagicMock(name='load_from_file',
                                        return_value=self)

    def __call__(self, **kwargs):
        """Mock call to return self."""
        self._kwargs = kwargs
        return self


class MockRunwayContext(RunwayContext):
    """Subclass Runway context object for tests."""

    def __init__(self,
                 env_name,
                 env_region,
                 env_root,
                 env_vars=None,
                 command=None):
        """Instantiate class."""
        super(MockRunwayContext, self).__init__(env_name=env_name or 'test',
                                                env_region=env_region or 'us-east-1',
                                                env_root=env_root,
                                                env_vars=env_vars,
                                                command=command)
        self._account_id = '123456789012'
        self._boto3_test_client = MutableMap()
        self._boto3_test_stubber = MutableMap()
        self._use_concurrent = True

    @property
    def account_id(self):
        """Override account_id."""
        return self._account_id

    @account_id.setter
    def account_id(self, value):
        """Set account_id."""
        self._account_id = value

    def add_stubber(self, service_name, region=None):
        """Add a stubber to context.

        Args:
            service_name (str): The name of a service, e.g. 's3' or 'ec2'.

        """
        key = '{}.{}'.format(service_name, region or self.env_region)

        self._boto3_test_client[key] = boto3.client(
            service_name,
            region_name=region or self.env_region,
            **self.boto3_credentials
        )
        self._boto3_test_stubber[key] = Stubber(
            self._boto3_test_client[key]
        )
        return self._boto3_test_stubber[key]

    def get_session(self, profile=None, region=None):
        """Wrap get_session to enable stubbing."""
        return MockBoto3Session(clients=self._boto3_test_client,
                                profile_name=profile,
                                region_name=region or self.env_region)

    @property
    def use_concurrent(self):
        """Override property of parent with something that can be set."""
        return self._use_concurrent

    @use_concurrent.setter
    def use_concurrent(self, value):
        """Override property of parent with something that can be set.

        Args:
            value (bool): New value for the attribute.

        """
        self._use_concurrent = value


class YamlLoader(object):
    """Load YAML files from a directory."""

    def __init__(self, root, load_class=None, load_type='default'):
        """Instantiate class.

        Args:
            root (Path): Root directory.
            load_class (Any): Class to use with load method.
            load_type (str): Contolls how content is passed to the load_class.

        """
        self.load_class = load_class
        self.load_type = load_type
        root.absolute()
        self.root = root

    def get(self, file_name):
        """Get raw YAML file contents.

        Args:
            file_name (str): Name of the file to load.

        Returns:
            Dict[str, Any]: Content of the file loaded by PyYAML.

        """
        if not file_name.endswith('.yml') or not file_name.endswith('.yaml'):
            file_name += '.yml'
        if sys.version_info.major > 2:
            content = (self.root / file_name).read_text()
        else:
            content = (self.root / file_name).read_text().decode()
        return yaml.safe_load(content)

    def load(self, file_name):
        """Load YAML file contents.

        Args:
            file_name (str): Name of the file to load.

        Returns:
            Any

        """
        if not self.load_class:
            raise ValueError('load_class must be set to use this method')
        if self.load_type == 'default':
            return self.load_class(self.get(file_name))
        if self.load_type == 'kwargs':
            return self.load_class(**self.get(file_name))
        raise ValueError('invalid load_type; "{}"'.format(self.load_type))


class YamlLoaderDeploymet(YamlLoader):
    """Load deployment YAML files from a directory."""

    def __init__(self, root):
        """Instantiate class.

        Args:
            root (Path): Root directory.

        """
        super(YamlLoaderDeploymet, self).__init__(root,
                                                  load_class=DeploymentDefinition)

    def load(self, file_name):
        """Load YAML file contents.

        Args:
            file_name (str): Name of the file to load.

        Returns:
            DeploymentDefinition

        """
        return self.load_class.from_list([self.get(file_name)])[0]



"""Test changing backends between local and local."""
from integration_tests.test_terraform.test_terraform import Terraform


class LocalToLocalBackend(Terraform):
    """Test changing between a specific and non-specific local backends."""

    TEST_NAME = __name__

    def deploy_backend(self, backend):
        """Deploy provider."""
        self.copy_template('{}-backend.tf'.format(backend))
        self.copy_runway('nos3')
        code, _stdout, _stderr = self.runway_cmd('deploy')
        return code

    def run(self):
        """Run tests."""
        self.clean()
        self.set_tf_version(11)

        assert self.deploy_backend('no') == 0,\
            '{}: "No local backend" failed'.format(self.TEST_NAME)
        # https://github.com/hashicorp/terraform/issues/17663
        assert self.deploy_backend('local') != 0,\
            '{}: "Local backend" failed'.format(self.TEST_NAME)

    def teardown(self):
        """Teardown any created resources."""
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        code, _stdout, _stderr = self.runway_cmd('destroy')
        assert code == 0, 'exit code should be zero'
        self.clean()



"""Runner for integration tests."""
from __future__ import print_function
import os
import logging

from .integration_test import IntegrationTest
from .util import (execute_tests, import_tests)


class Runner(object):
    """Runner for all integration tests."""

    def __init__(self, test_to_run=None, use_abs=False):
        """Initialize object.

        Args:
            test_to_run: Name of a test to run.
            package: Used as the anchor for relative imports

        """
        self._use_abs = use_abs
        self._tests_imported = False
        if os.environ.get('DEBUG'):
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.INFO)
        self.logger = logging.getLogger('testsuite')
        self.working_dir = os.path.abspath(os.path.dirname(__file__))
        if test_to_run:
            self.test_to_run = 'test_{0}/test_{0}'.format(test_to_run)
        else:
            self.test_to_run = 'test_*/test_*'

    @property
    def available_tests(self):
        if not self._tests_imported:
            import_tests(self.logger, self.working_dir, self.test_to_run, self._use_abs)
            self._tests_imported = True
        return IntegrationTest.__subclasses__()

    def run_tests(self):
        """Run all integration tests."""
        return execute_tests([test(self.logger) for test in self.available_tests],
                             self.logger)

    def main(self):
        """Import and run tests."""
        errs = self.run_tests()
        if errs > 0:
            self.logger.error('Tests failed; Check logs.')
        return 1 if errs > 0 else 0



"""Client for handling retrieval of JWKS signing keys."""
import json
import logging
import urllib

# pylint: disable=relative-beyond-top-level
from .utils import rsa_public_key_to_pem


def is_signing_key(key):
    """Filter to determine if this is a signing key.

    Keyword Args:
        key (Dict[str, str]): The key
    """
    if key.get('kty', '') != 'RSA':
        return False
    if not key.get('kid', None):
        return False
    if key.get('use', '') != 'sig':
        return False
    return key.get('x5c') or (key.get('n') and key.get('e'))


class JwksClient(object):
    """Client responsible for retrieval of JWKS signing keys."""

    def __init__(self, options=None):
        """Initialize.

        Keyword Args:
            options (Optional[Dict[str, str]]): Options for the client
        """
        self.options = options
        self.logger = logging.getLogger('__file__')

    def get_keys(self):
        """Retrieve the keys from the JWKS endpoint."""
        self.logger.info('Fetching keys from %s', self.options.get('jwks_uri'))

        try:
            request = urllib.request.urlopen(self.options.get('jwks_uri'))  # noqa pylint: disable=no-member
            data = json.loads(
                request.read().decode(
                    request.info().get_param('charset') or 'utf-8'
                )
            )
            keys = data['keys']
            self.logger.info("Keys: %s", keys)
            return keys
        # pylint: disable=broad-except
        except Exception as err:
            self.logger.info('Failure: ConnectionError')
            self.logger.info(err)
            return {}

    def get_signing_key(self, kid):
        """Given a specific key id (kid) retrieve the signing key associated.

        Keyword Args:
            kid (str): The key id of the signing key
        """
        self.logger.info('Fetching signing key for %s', kid)

        keys = self.get_signing_keys()
        try:
            key = next(x for x in keys if x.get('kid') == kid)
            return key
        except StopIteration:
            raise Exception('Was not able to locate a key with kid %s' % kid)

    def get_signing_keys(self):
        """Given a set of keys find all that are signing keys."""
        keys = self.get_keys()

        if not keys:
            raise Exception('The JWKS endpoint did not contain any keys')

        jwks = []
        for key in filter(is_signing_key, keys):
            jwks.append(self.create_jwk(key))

        if not jwks:
            raise Exception("The JWKS endpoint did not contain any signing keys")

        self.logger.info("Signing Keys: %s", jwks)
        return jwks

    def create_jwk(self, key):
        """Create the JSON Web Key.

        Keyword Args:
            key (dict): The Retrieved signing key
        """
        jwk = {
            'kid': key.get('kid'),
            'nbf': key.get('nbf')
        }

        if key.get('x5c'):
            # @TODO: Support certificate chains. Review library here:
            # https://github.com/auth0/node-jwks-rsa/blob/master/src/JwksClient.js#L87
            self.logger.info('X5C')
        else:
            try:
                jwk['rsaPublicKey'] = rsa_public_key_to_pem(
                    key.get('n'),
                    key.get('e')
                )
            # pylint: disable=broad-except
            except Exception as err:
                self.logger.error(err)
                jwk['rsaPublicKey'] = None
        return jwk



"""Tests for variables."""
# pylint: disable=protected-access,unused-argument
from unittest import TestCase

from mock import MagicMock
from troposphere import s3

from runway.cfngin.blueprints.variables.types import TroposphereType
from runway.cfngin.exceptions import UnresolvedVariable
from runway.cfngin.lookups import register_lookup_handler
from runway.cfngin.stack import Stack
from runway.util import MutableMap
from runway.variables import Variable

from .cfngin.factories import generate_definition

VALUE = {
    'test': 'success',
    'what': 'test',
    'bool_val': False,
    'dict_val': {'test': 'success'},
    'list_val': ['success']
}
CONTEXT = MutableMap(**{
    'env_vars': VALUE
})


class TestCfnginVariables(TestCase):
    """Tests for runway.cfngin.variables."""

    def setUp(self):
        """Run before tests."""
        self.provider = MagicMock()
        self.context = MagicMock()

    def test_variable_replace_no_lookups(self):
        """Test variable replace no lookups."""
        var = Variable("Param1", "2")
        self.assertEqual(var.value, "2")

    def test_variable_replace_simple_lookup(self):
        """Test variable replace simple lookup."""
        var = Variable("Param1", "${output fakeStack::FakeOutput}")
        var._value._resolve("resolved")
        self.assertEqual(var.value, "resolved")

    def test_variable_resolve_simple_lookup(self):
        """Test variable resolve simple lookup."""
        stack = Stack(
            definition=generate_definition("vpc", 1),
            context=self.context)
        stack.set_outputs({
            "FakeOutput": "resolved",
            "FakeOutput2": "resolved2",
        })

        self.context.get_stack.return_value = stack

        var = Variable("Param1", "${output fakeStack::FakeOutput}")
        var.resolve(self.context, self.provider)
        self.assertTrue(var.resolved)
        self.assertEqual(var.value, "resolved")

    def test_variable_resolve_default_lookup_empty(self):
        """Test variable resolve default lookup empty."""
        var = Variable("Param1", "${default fakeStack::}")
        var.resolve(self.context, self.provider)
        self.assertTrue(var.resolved)
        self.assertEqual(var.value, "")

    def test_variable_replace_multiple_lookups_string(self):
        """Test variable replace multiple lookups string."""
        var = Variable(
            "Param1",
            "url://"  # 0
            "${output fakeStack::FakeOutput}"  # 1
            "@"  # 2
            "${output fakeStack::FakeOutput2}",  # 3
        )
        var._value[1]._resolve("resolved")
        var._value[3]._resolve("resolved2")
        self.assertEqual(var.value, "url://resolved@resolved2")

    def test_variable_resolve_multiple_lookups_string(self):
        """Test variable resolve multiple lookups string."""
        var = Variable(
            "Param1",
            "url://${output fakeStack::FakeOutput}@"
            "${output fakeStack::FakeOutput2}",
        )

        stack = Stack(
            definition=generate_definition("vpc", 1),
            context=self.context)
        stack.set_outputs({
            "FakeOutput": "resolved",
            "FakeOutput2": "resolved2",
        })

        self.context.get_stack.return_value = stack
        var.resolve(self.context, self.provider)
        self.assertTrue(var.resolved)
        self.assertEqual(var.value, "url://resolved@resolved2")

    def test_variable_replace_no_lookups_list(self):
        """Test variable replace no lookups list."""
        var = Variable("Param1", ["something", "here"])
        self.assertEqual(var.value, ["something", "here"])

    def test_variable_replace_lookups_list(self):
        """Test variable replace lookups list."""
        value = ["something",  # 0
                 "${output fakeStack::FakeOutput}",  # 1
                 "${output fakeStack::FakeOutput2}"  # 2
                 ]
        var = Variable("Param1", value)

        var._value[1]._resolve("resolved")
        var._value[2]._resolve("resolved2")
        self.assertEqual(var.value, ["something", "resolved", "resolved2"])

    def test_variable_replace_lookups_dict(self):
        """Test variable replace lookups dict."""
        value = {
            "something": "${output fakeStack::FakeOutput}",
            "other": "${output fakeStack::FakeOutput2}",
        }
        var = Variable("Param1", value)
        var._value["something"]._resolve("resolved")
        var._value["other"]._resolve("resolved2")
        self.assertEqual(var.value, {"something": "resolved", "other":
                                     "resolved2"})

    def test_variable_replace_lookups_mixed(self):
        """Test variable replace lookups mixed."""
        value = {
            "something": [
                "${output fakeStack::FakeOutput}",
                "other",
            ],
            "here": {
                "other": "${output fakeStack::FakeOutput2}",
                "same": "${output fakeStack::FakeOutput}",
                "mixed": "something:${output fakeStack::FakeOutput3}",
            },
        }
        var = Variable("Param1", value)
        var._value["something"][0]._resolve("resolved")
        var._value["here"]["other"]._resolve("resolved2")
        var._value["here"]["same"]._resolve("resolved")
        var._value["here"]["mixed"][1]._resolve("resolved3")
        self.assertEqual(var.value, {
            "something": [
                "resolved",
                "other",
            ],
            "here": {
                "other": "resolved2",
                "same": "resolved",
                "mixed": "something:resolved3",
            },
        })

    def test_variable_resolve_nested_lookup(self):
        """Test variable resolve nested lookup."""
        stack = Stack(
            definition=generate_definition("vpc", 1),
            context=self.context)
        stack.set_outputs({
            "FakeOutput": "resolved",
            "FakeOutput2": "resolved2",
        })

        def mock_handler(value, context, provider, **kwargs):
            return "looked up: {}".format(value)

        register_lookup_handler("lookup", mock_handler)
        self.context.get_stack.return_value = stack
        var = Variable(
            "Param1",
            "${lookup ${lookup ${output fakeStack::FakeOutput}}}",
        )
        var.resolve(self.context, self.provider)
        self.assertTrue(var.resolved)
        self.assertEqual(var.value, "looked up: looked up: resolved")

    def test_troposphere_type_no_from_dict(self):
        """Test troposphere type no from dict."""
        with self.assertRaises(ValueError):
            TroposphereType(object)

        with self.assertRaises(ValueError):
            TroposphereType(object, many=True)

    def test_troposphere_type_create(self):
        """Test troposphere type create."""
        troposphere_type = TroposphereType(s3.Bucket)
        created = troposphere_type.create(
            {"MyBucket": {"BucketName": "test-bucket"}})
        self.assertTrue(isinstance(created, s3.Bucket))
        self.assertTrue(created.properties["BucketName"], "test-bucket")

    def test_troposphere_type_create_multiple(self):
        """Test troposphere type create multiple."""
        troposphere_type = TroposphereType(s3.Bucket, many=True)
        created = troposphere_type.create({
            "FirstBucket": {"BucketName": "test-bucket"},
            "SecondBucket": {"BucketName": "other-test-bucket"},
        })
        self.assertTrue(isinstance(created, list))


class TestRunwayVariables(TestCase):
    """Test for variable resolution."""

    def test_value_simple_str(self):
        """Test value for a simple string without lookups."""
        var = Variable('test', 'success')

        self.assertTrue(var.resolved, msg='when no lookup is used, it should '
                        'be automatically marked as resolved')
        self.assertEqual(var.value, 'success')

    def test_value_simple_str_lookup(self):
        """Test value for simple str lookup."""
        var = Variable('test', '${env test}', 'runway')

        self.assertFalse(var.resolved)

        var.resolve(CONTEXT)

        self.assertTrue(var.resolved)
        self.assertEqual(var.value, VALUE['test'])

    def test_value_complex_str(self):
        """Multiple lookups should be usable within a single string."""
        var = Variable('test', 'the ${env what} was ${env test}ful', 'runway')
        var.resolve(CONTEXT)

        self.assertEqual(var.value, 'the {} was {}ful'.format(VALUE['what'],
                                                              VALUE['test']))

    def test_value_nested_str(self):
        """Variable lookups should be resolvable within each other."""
        var = Variable('test', '${env ${env what}}', 'runway')
        var.resolve(CONTEXT)

        self.assertEqual(var.value, VALUE['test'])

    def test_value_lookup_in_dict(self):
        """Variable lookups should be resolvable when used in a dict."""
        var = Variable('test', {'my_dict': '${env test}'}, 'runway')
        var.resolve(CONTEXT)

        self.assertEqual(var.value, {'my_dict': VALUE['test']})

    def test_value_lookup_in_list(self):
        """Variable lookups should be resolvable when used in a list."""
        var = Variable('test', ['${env test}'], 'runway')
        var.resolve(CONTEXT)

        self.assertEqual(var.value, [VALUE['test']])

    def test_value_lookup_to_bool(self):
        """Variable lookups should be resolvable to a bool."""
        var = Variable('test', '${env bool_val}', 'runway')
        var.resolve(CONTEXT)

        self.assertFalse(var.value)

    def test_value_lookup_to_dict(self):
        """Variable lookups should be resolvable to a dict value."""
        var = Variable('test', '${env dict_val}', 'runway')
        var.resolve(CONTEXT)

        self.assertEqual(var.value, VALUE['dict_val'])

    def test_value_lookup_to_list(self):
        """Variable lookups should be resolvable to a list value."""
        var = Variable('test', '${env list_val}', 'runway')
        var.resolve(CONTEXT)

        self.assertEqual(var.value, VALUE['list_val'])

    def test_value_unresolved(self):
        """Should raise `UnresolvedVariable`."""
        var = Variable('test', '${env test}', 'runway')

        with self.assertRaises(UnresolvedVariable):
            print(var.value)



"""Pytest fixtures and plugins."""
# pylint: disable=redefined-outer-name
import logging
import os
import sys
from typing import Dict

import pytest
import yaml

from runway.config import Config

from .factories import (MockCFNginContext, MockRunwayConfig, MockRunwayContext,
                        YamlLoader, YamlLoaderDeploymet)

if sys.version_info.major > 2:  # TODO remove after droping python 2
    from pathlib import Path  # pylint: disable=E
else:
    from pathlib2 import Path  # pylint: disable=E

LOG = logging.getLogger(__name__)
TEST_ROOT = Path(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture(scope='session', autouse=True)
def aws_credentials():
    # type: () -> Dict[str, str]
    """Handle change in https://github.com/spulec/moto/issues/1924.

    Ensure AWS SDK finds some (bogus) credentials in the environment and
    doesn't try to use other providers.

    """
    overrides = {
        'AWS_ACCESS_KEY_ID': 'testing',
        'AWS_SECRET_ACCESS_KEY': 'testing',
        'AWS_DEFAULT_REGION': 'us-east-1'
    }
    saved_env = {}
    for key, value in overrides.items():
        LOG.info('Overriding env var: %s=%s', key, value)
        saved_env[key] = os.environ.get(key, None)
        os.environ[key] = value

    yield

    for key, value in saved_env.items():
        LOG.info('Restoring saved env var: %s=%s', key, value)
        if value is None:
            os.environ.pop(key, None)  # handle key missing
        else:
            os.environ[key] = value

    saved_env.clear()


@pytest.fixture(scope='session', autouse=True)
def sanitize_environment():
    # type: () -> None
    """Remove variables from the environment that could interfere with tests."""
    env_vars = ['CI', 'DEBUG', 'DEPLOY_ENVIRONMENT', 'CFNGIN_STACK_POLL_TIME',
                'RUNWAY_MAX_CONCURRENT_MODULES',
                'RUNWAY_MAX_CONCURRENT_REGIONS']
    for var in env_vars:
        os.environ.pop(var, None)


@pytest.fixture(scope='package')
def fixture_dir():
    # type: () -> str
    """Path to the fixture directory."""
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        'fixtures')
    return path


@pytest.fixture(scope='module')
def fx_config():
    """Return YAML loader for config fixtures."""
    return YamlLoader(TEST_ROOT / 'fixtures' / 'configs',
                      load_class=Config,
                      load_type='kwargs')


@pytest.fixture(scope='function')
def fx_deployments():
    """Return YAML loader for deployment fixtures."""
    return YamlLoaderDeploymet(TEST_ROOT / 'fixtures' / 'deployments')


@pytest.fixture(scope='module')
def yaml_fixtures(request, fixture_dir):
    """Load test fixture yaml files.

    Uses a list of file paths within the fixture directory loaded from the
    `YAML_FIXTURES` variable of the module.

    """
    file_paths = getattr(request.module, 'YAML_FIXTURES', [])
    result = {}
    for file_path in file_paths:
        with open(os.path.join(fixture_dir, file_path)) as _file:
            data = _file.read()
            result[file_path] = yaml.safe_load(data)
    return result


@pytest.fixture(scope='function')
def cfngin_context(runway_context):
    """Create a mock CFNgin context object."""
    return MockCFNginContext(environment={},
                             boto3_credentials=runway_context.boto3_credentials,
                             region=runway_context.env_region)


@pytest.fixture
def patch_time(monkeypatch):
    """Patch built-in time object."""
    monkeypatch.setattr('time.sleep', lambda s: None)


@pytest.fixture(scope='function')
def patch_runway_config(request, monkeypatch, runway_config):
    """Patch Runway config and return a mock config object."""
    patch_path = getattr(request.module, 'PATCH_RUNWAY_CONFIG', None)
    if patch_path:
        monkeypatch.setattr(patch_path, runway_config)
    return runway_config


@pytest.fixture(scope='function')
def runway_config():
    """Create a mock runway config object."""
    return MockRunwayConfig()


@pytest.fixture(scope='function')
def runway_context(request):
    """Create a mock Runway context object."""
    env_vars = {
        'AWS_REGION': getattr(request.module, 'AWS_REGION', 'us-east-1'),
        'DEFAULT_AWS_REGION': getattr(request.module, 'AWS_REGION', 'us-east-1'),
        'DEPLOY_ENVIRONMET': getattr(request.module, 'DEPLOY_ENVIRONMET', 'test')
    }
    creds = {
        'AWS_ACCESS_KEY_ID': 'test_access_key',
        'AWS_SECRET_ACCESS_KEY': 'test_secret_key',
        'AWS_SESSION_TOKEN': 'test_session_token'
    }
    env_vars.update(getattr(request.module, 'AWS_CREDENTIALS', creds))
    env_vars.update(getattr(request.module, 'ENV_VARS', {}))
    return MockRunwayContext(env_name='test',
                             env_region='us-east-1',
                             env_root=os.getcwd(),
                             env_vars=env_vars,
                             command='test')



"""Self Destruction Lambda."""
from typing import Any, Dict, Union  # pylint: disable=unused-import

import boto3


def handler(event,  # type: Dict
            _context  # type: Dict
           ):  # noqa: E124
    # type: (...) -> Union[Dict[str, Any], bool]
    """Self destruct the stack.

    Execute self destruction of the stack this lambda is a part of.

    Args:
        event (dict): The AWS Step Function execution event
    """
    data = event.get('SelfDestruct')
    cfn_client = boto3.client('cloudformation')

    if not data:
        return False

    deleted_stack = cfn_client.delete_stack(StackName=data.get('StackName'))

    return {'deleted_stack': deleted_stack}



"""Tests for terraform module."""
# pylint: disable=no-self-use,unused-argument
import sys
from contextlib import contextmanager
from datetime import datetime

import boto3
import pytest
from botocore.stub import Stubber
from mock import patch

from runway.module.terraform import (TerraformBackendConfig, TerraformOptions,
                                     update_env_vars_with_tf_var_values)


@contextmanager
def does_not_raise():
    """Use for conditional pytest.raises when using parametrize."""
    yield


def test_update_env_vars_with_tf_var_values():
    """Test update_env_vars_with_tf_var_values."""
    result = update_env_vars_with_tf_var_values({}, {'foo': 'bar',
                                                     'list': ['foo', 1, True],
                                                     'map': sorted({  # python 2
                                                         'one': 'two',
                                                         'three': 'four'
                                                     })})
    expected = {
        'TF_VAR_foo': 'bar',
        'TF_VAR_list': '["foo", 1, true]',
        'TF_VAR_map': '{ one = "two", three = "four" }'
    }

    assert sorted(result) == sorted(expected)  # sorted() needed for python 2


class TestTerraformOptions(object):
    """Test runway.module.terraform.TerraformOptions."""

    @pytest.mark.parametrize('config', [
        ({'args': ['-key=val']}),
        ({'args': {'apply': ['-key=apply']}}),
        ({'args': {'init': ['-key=init']}}),
        ({'args': {'plan': ['-key=plan']}}),
        ({'args': {'apply': ['-key=apply'], 'init': ['-key=init']}}),
        ({'args': {'apply': ['-key=apply'],
                   'init': ['-key=init'],
                   'plan': ['-key=plan']}}),
        ({'terraform_backend_config': {'bucket': 'foo',
                                       'dynamodb_table': 'bar',
                                       'region': 'us-west-2'}}),
        ({'terraform_backend_config': {'region': 'us-west-2'},
          'terraform_backend_cfn_outputs': {'bucket': 'foo',
                                            'dynamodb_table': 'bar'}}),
        ({'terraform_backend_config': {'region': 'us-west-2'},
          'terraform_backend_ssm_params': {'bucket': 'foo',
                                           'dynamodb_table': 'bar'}}),
        ({'terraform_version': '0.11.6'}),
        ({'terraform_version': {'test': '0.12', 'prod': '0.11.6'}}),  # deprecated
        ({'args': {'apply': ['-key=apply'],
                   'init': ['-key=init'],
                   'plan': ['-key=plan']},
          'terraform_backend_config': {'region': 'us-west-2'},
          'terraform_backend_ssm_params': {'bucket': 'foo',
                                           'dynamodb_table': 'bar'},
          'terraform_version': {'test': '0.12', 'prod': '0.11.6'}}),  # deprecated
        ({'args': {'apply': ['-key=apply'],
                   'init': ['-key=init'],
                   'plan': ['-key=plan']},
          'terraform_backend_config': {'bucket': 'foo',
                                       'dynamodb_table': 'bar',
                                       'region': 'us-west-2'},
          'terraform_version': '0.11.6'}),
        ({'args': ['-key=val'],  # deprecated
          'terraform_backend_config': {'test': {'bucket': 'foo',
                                                'dynamodb_table': 'bar'},
                                       'prod': {'bucket': 'invalid',
                                                'dynamodb_table': 'invalid'}},
          'terraform_version': {'test': '0.12', 'prod': '0.11.6'}})
    ])
    @patch('runway.module.terraform.TerraformBackendConfig.parse')
    def test_parse(self, mock_backend, config, monkeypatch, runway_context):
        """Test parse."""
        mock_backend.return_value = 'successfully parsed backend'

        if sys.version_info.major < 3:  # python 2 support
            @staticmethod
            def assert_resolve_version_kwargs(context, terraform_version=None,
                                              **_):
                """Assert args passed to the method during parse."""
                assert config.get('terraform_version') == terraform_version
                return 'successfully resolved version'
        else:
            def assert_resolve_version_kwargs(context, terraform_version=None,
                                              **_):
                """Assert args passed to the method during parse."""
                assert config.get('terraform_version') == terraform_version
                return 'successfully resolved version'

        monkeypatch.setattr(TerraformOptions, 'resolve_version',
                            assert_resolve_version_kwargs)

        result = TerraformOptions.parse(context=runway_context,
                                        path='./', **config)

        if isinstance(config.get('args'), list):
            assert result.args['apply'] == config['args']
            assert result.args['init'] == []
            assert result.args['plan'] == []
        elif isinstance(config.get('args'), dict):
            assert result.args['apply'] == config['args'].get('apply', [])
            assert result.args['init'] == config['args'].get('init', [])
            assert result.args['plan'] == config['args'].get('plan', [])
        assert result.backend_config == 'successfully parsed backend'
        assert result.version == 'successfully resolved version'
        mock_backend.assert_called_once_with(runway_context, './', **config)

    @pytest.mark.parametrize('terraform_version, expected, exception',
                             [('0.11.6', '0.11.6', does_not_raise()),
                              ({'test': '0.12', 'prod': '0.11.6'},  # deprecated
                               '0.12', does_not_raise()),
                              ({'*': '0.11.6', 'test': '0.12'},  # deprecated
                               '0.12', does_not_raise()),
                              ({'*': '0.11.6', 'prod': '0.12'},  # deprecated
                               '0.11.6', does_not_raise()),
                              ({'prod': '0.11.6'}, None,  # deprecated
                               does_not_raise()),
                              (None, None, does_not_raise()),
                              (13, None, pytest.raises(TypeError))])
    def test_resolve_version(self, runway_context, terraform_version,
                             expected, exception):
        """Test resolve_version."""
        config = {'something': None}
        if terraform_version:
            config['terraform_version'] = terraform_version
        with exception:
            assert TerraformOptions.resolve_version(runway_context,
                                                    **config) == expected


class TestTerraformBackendConfig(object):
    """Test runway.module.terraform.TerraformBackendConfig."""

    @pytest.mark.parametrize('input_data, expected_items', [
        ({}, []),
        ({'bucket': 'test-bucket'}, ['bucket=test-bucket']),
        ({'dynamodb_table': 'test-table'}, ['dynamodb_table=test-table']),
        ({'region': 'us-east-1'}, ['region=us-east-1']),
        ({'bucket': 'test-bucket', 'dynamodb_table': 'test-table'},
         ['bucket=test-bucket', 'dynamodb_table=test-table']),
        ({'bucket': 'test-bucket', 'dynamodb_table': 'test-table',
          'region': 'us-east-1'},
         ['bucket=test-bucket', 'dynamodb_table=test-table',
          'region=us-east-1']),
        ({'bucket': 'test-bucket', 'dynamodb_table': 'test-table',
          'region': 'us-east-1', 'filename': 'anything'},
         ['bucket=test-bucket', 'dynamodb_table=test-table',
          'region=us-east-1'])
    ])
    def test_init_args(self, input_data, expected_items):
        """Test init_args."""
        expected = []
        for i in expected_items:
            expected.extend(['-backend-config', i])
        assert TerraformBackendConfig(**input_data).init_args == expected

    @pytest.mark.parametrize('kwargs, stack_info,expected', [
        ({'bucket': 'tf-state::BucketName',
          'dynamodb_table': 'tf-state::TableName'},
         {'tf-state': {'BucketName': 'test-bucket', 'TableName': 'test-table'}},
         {'bucket': 'test-bucket', 'dynamodb_table': 'test-table'}),
        ({}, {}, {})
    ])
    def test_resolve_cfn_outputs(self, kwargs, stack_info, expected):
        """Test resolve_cfn_outputs."""
        client = boto3.client('cloudformation')
        stubber = Stubber(client)
        for stack, outputs in stack_info.items():
            for key, val in outputs.items():
                stubber.add_response(
                    'describe_stacks',
                    {
                        'Stacks': [{
                            'StackName': stack,
                            'CreationTime': datetime.now(),
                            'StackStatus': 'CREATE_COMPLETE',
                            'Outputs': [{
                                'OutputKey': key,
                                'OutputValue': val
                            }]
                        }]
                    }
                )
        with stubber:
            assert TerraformBackendConfig.resolve_cfn_outputs(
                client, **kwargs
            ) == expected
        stubber.assert_no_pending_responses()

    @pytest.mark.parametrize('kwargs, parameters, expected', [
        ({'bucket': '/some/param/key', 'dynamodb_table': 'foo'},
         [{'name': '/some/param/key', 'value': 'test-bucket'},
          {'name': 'foo', 'value': 'test-table'}],
         {'bucket': 'test-bucket', 'dynamodb_table': 'test-table'}),
        ({}, {}, {})
    ])
    @pytest.mark.skipif(sys.version_info.major < 3,
                        reason='python 2 dict handling prevents this from '
                        'reliably passing')
    def test_resolve_ssm_params(self, caplog, kwargs, parameters, expected):
        """Test resolve_ssm_params."""
        # this test is not compatable with python 2 due to how it handles dicts
        caplog.set_level('WARNING', logger='runway')

        client = boto3.client('ssm')
        stubber = Stubber(client)

        for param in parameters:
            stubber.add_response(
                'get_parameter',
                {
                    'Parameter': {
                        'Name': param['name'],
                        'Value': param['value'],
                        'LastModifiedDate': datetime.now()
                    }
                },
                {'Name': param['name'], 'WithDecryption': True}
            )

        with stubber:
            assert TerraformBackendConfig.resolve_ssm_params(
                client, **kwargs
            ) == expected
        stubber.assert_no_pending_responses()
        assert 'deprecated' in caplog.records[0].msg

    def test_gen_backend_tfvars_filenames(self):
        """Test gen_backend_tfvars_filenames."""
        expected = ['backend-test-us-east-1.tfvars',
                    'backend-test.tfvars',
                    'backend-us-east-1.tfvars',
                    'backend.tfvars']

        assert TerraformBackendConfig.gen_backend_tfvars_filenames(
            'test', 'us-east-1'
        ) == expected

    @pytest.mark.parametrize('filename, expected', [
        ('backend-test-us-east-1.tfvars', 'backend-test-us-east-1.tfvars'),
        ('backend-test.tfvars', 'backend-test.tfvars'),
        ('backend-us-east-1.tfvars', 'backend-us-east-1.tfvars'),
        ('backend.tfvars', 'backend.tfvars'),
        ('something-backend.tfvars', None),
        (['backend-test-us-east-1.tfvars', 'backend.tfvars'],
         'backend-test-us-east-1.tfvars')
    ])
    def test_get_backend_tfvars_file(self, tmp_path, filename, expected):
        """Test get_backend_tfvars_file."""
        if isinstance(filename, list):
            for name in filename:
                (tmp_path / name).touch()
        else:
            (tmp_path / filename).touch()
        # TODO remove conversion of path to str when dripping python 2
        assert TerraformBackendConfig.get_backend_tfvars_file(
            str(tmp_path), 'test', 'us-east-1'
        ) == expected

    @pytest.mark.parametrize('config, expected_region', [
        ({'terraform_backend_config': {'bucket': 'foo',
                                       'dynamodb_table': 'bar',
                                       'region': 'us-west-2'}}, 'us-west-2'),
        ({'terraform_backend_config': {'region': 'us-west-2'},
          'terraform_backend_cfn_outputs': {'bucket': 'foo',
                                            'dynamodb_table': 'bar'}},
         'us-west-2'),
        ({'terraform_backend_config': {'region': 'us-west-2'},
          'terraform_backend_ssm_params': {'bucket': 'foo',  # deprecated
                                           'dynamodb_table': 'bar'}},
         'us-west-2'),
        ({'terraform_backend_config': {'bucket': 'foo',
                                       'dynamodb_table': 'bar'}},
         'us-east-1'),
        ({'terraform_backend_cfn_outputs': {'bucket': 'foo',
                                            'dynamodb_table': 'bar'}},
         'us-east-1'),
        ({'terraform_backend_ssm_params': {'bucket': 'foo',  # deprecated
                                           'dynamodb_table': 'bar'}},
         'us-east-1'),
        ({'terraform_backend_cfn_outputs': {'bucket': 'foo',
                                            'dynamodb_table': 'bar'}},
         'us-east-1'),
        ({'terraform_backend_ssm_params': {'bucket': 'foo',  # deprecated
                                           'dynamodb_table': 'bar'}},
         'us-east-1'),
        ({'terraform_backend_cfn_outputs': {'bucket': 'foo'},
          'terraform_backend_ssm_params': {'dynamodb_table': 'bar'}},  # deprecated
         'us-east-1'),
        ({'terraform_backend_config': {'bucket': 'nope',
                                       'dynamodb_table': 'nope',
                                       'region': 'us-west-2'},
          'terraform_backend_cfn_outputs': {'bucket': 'foo',
                                            'dynamodb_table': 'bar'}},
         'us-west-2'),
        ({'terraform_backend_config': {'bucket': 'nope',
                                       'dynamodb_table': 'nope',
                                       'region': 'us-west-2'},
          'terraform_backend_ssm_params': {'bucket': 'foo',  # deprecated
                                           'dynamodb_table': 'bar'}},
         'us-west-2'),
        ({'terraform_backend_cfn_outputs': {'bucket': 'nope',
                                            'dynamodb_table': 'nope'},
          'terraform_backend_ssm_params': {'bucket': 'foo',  # deprecated
                                           'dynamodb_table': 'bar'}},
         'us-east-1'),
        ({'terraform_backend_config': {'test': {'bucket': 'foo',  # deprecated
                                                'dynamodb_table': 'bar'},
                                       'prod': {'bucket': 'invalid',
                                                'dynamodb_table': 'invalid'}}},
         'us-east-1')
    ])
    def test_parse(self, monkeypatch, runway_context, config, expected_region):
        """Test parse."""
        runway_context.add_stubber('cloudformation', expected_region)
        runway_context.add_stubber('ssm', expected_region)

        if sys.version_info.major < 3:  # python 2 support
            @staticmethod
            def assert_cfn_kwargs(client, **kwargs):
                """Assert args passed to the method during parse."""
                assert kwargs == config.get('terraform_backend_cfn_outputs')
                return kwargs

            @staticmethod
            def assert_ssm_kwargs(client, **kwargs):
                """Assert args passed to the method during parse."""
                assert kwargs == config.get('terraform_backend_ssm_params')
                return kwargs

            @classmethod
            def assert_get_backend_tfvars_file_args(_, path, env_name, env_region):
                """Assert args passed to the method during parse."""
                assert path == './'
                assert env_name == 'test'
                assert env_region == 'us-east-1'
                return 'success'
        else:
            def assert_cfn_kwargs(client, **kwargs):
                """Assert args passed to the method during parse."""
                assert kwargs == config.get('terraform_backend_cfn_outputs')
                return kwargs

            def assert_ssm_kwargs(client, **kwargs):
                """Assert args passed to the method during parse."""
                assert kwargs == config.get('terraform_backend_ssm_params')
                return kwargs

            def assert_get_backend_tfvars_file_args(path, env_name, env_region):
                """Assert args passed to the method during parse."""
                assert path == './'
                assert env_name == 'test'
                assert env_region == 'us-east-1'
                return 'success'

        monkeypatch.setattr(TerraformBackendConfig, 'resolve_cfn_outputs',
                            assert_cfn_kwargs)
        monkeypatch.setattr(TerraformBackendConfig,
                            'resolve_ssm_params',
                            assert_ssm_kwargs)
        monkeypatch.setattr(TerraformBackendConfig, 'get_backend_tfvars_file',
                            assert_get_backend_tfvars_file_args)

        result = TerraformBackendConfig.parse(runway_context, './', **config)

        assert result.bucket == 'foo'
        assert result.dynamodb_table == 'bar'
        assert result.region == expected_region
        assert result.filename == 'success'



"""Test runway.module.serverless."""
# pylint: disable=no-self-use,unused-argument
import logging
import subprocess
import sys

import pytest
import yaml
from mock import ANY, MagicMock, patch

from runway.module.serverless import (Serverless, ServerlessOptions,
                                      gen_sls_config_files)

from ..factories import MockProcess


@pytest.mark.usefixtures('patch_module_npm')
class TestServerless(object):
    """Test runway.module.serverless.Serverless."""

    def test_cli_args(self, runway_context):
        """Test cli_args."""
        obj = Serverless(runway_context, './tests')

        assert obj.cli_args == ['--region', runway_context.env_region,
                                '--stage', runway_context.env_name]

        runway_context.env_vars['DEBUG'] = '1'
        assert obj.cli_args == ['--region', runway_context.env_region,
                                '--stage', runway_context.env_name,
                                '--verbose']

    def test_deploy(self, monkeypatch, runway_context):
        """Test deploy."""
        # pylint: disable=no-member
        monkeypatch.setattr(Serverless, 'extend_serverless_yml', MagicMock())
        monkeypatch.setattr(Serverless, 'sls_deploy', MagicMock())
        obj = Serverless(runway_context, './tests')

        monkeypatch.setattr(Serverless, 'skip', True)
        assert not obj.deploy()
        obj.extend_serverless_yml.assert_not_called()
        obj.sls_deploy.assert_not_called()

        monkeypatch.setattr(Serverless, 'skip', False)
        monkeypatch.setattr(obj.options, 'extend_serverless_yml', True)
        assert not obj.deploy()
        obj.extend_serverless_yml.assert_called_once_with(obj.sls_deploy)
        obj.sls_deploy.assert_not_called()

        monkeypatch.setattr(obj.options, 'extend_serverless_yml', False)
        assert not obj.deploy()
        obj.extend_serverless_yml.assert_called_once()
        obj.sls_deploy.assert_called_once_with()

    def test_destroy(self, monkeypatch, runway_context):
        """Test destroy."""
        # pylint: disable=no-member
        monkeypatch.setattr(Serverless, 'extend_serverless_yml', MagicMock())
        monkeypatch.setattr(Serverless, 'sls_remove', MagicMock())
        obj = Serverless(runway_context, './tests')

        monkeypatch.setattr(Serverless, 'skip', True)
        assert not obj.destroy()
        obj.extend_serverless_yml.assert_not_called()
        obj.sls_remove.assert_not_called()

        monkeypatch.setattr(Serverless, 'skip', False)
        monkeypatch.setattr(obj.options, 'extend_serverless_yml', True)
        assert not obj.destroy()
        obj.extend_serverless_yml.assert_called_once_with(obj.sls_remove)
        obj.sls_remove.assert_not_called()

        monkeypatch.setattr(obj.options, 'extend_serverless_yml', False)
        assert not obj.destroy()
        obj.extend_serverless_yml.assert_called_once()
        obj.sls_remove.assert_called_once_with()

    def test_env_file(self, runway_context, tmp_path):
        """Test env_file.

        Testing the precedence of each path, create the files in order from
        lowerst to highest. After creating the file, the property's value
        is checked then cleared since the value is cached after the first
        time it is resolved.

        """
        env_dir = tmp_path / 'env'
        env_dir.mkdir()
        obj = Serverless(runway_context, tmp_path)
        assert not obj.env_file
        del obj.env_file

        config_test_json = tmp_path / 'config-test.json'
        config_test_json.touch()
        assert obj.env_file == config_test_json
        del obj.env_file

        env_test_json = env_dir / 'test.json'
        env_test_json.touch()
        assert obj.env_file == env_test_json
        del obj.env_file

        config_test_us_east_1_json = tmp_path / 'config-test-us-east-1.json'
        config_test_us_east_1_json.touch()
        assert obj.env_file == config_test_us_east_1_json
        del obj.env_file

        env_test_us_east_1_json = env_dir / 'test-us-east-1.json'
        env_test_us_east_1_json.touch()
        assert obj.env_file == env_test_us_east_1_json
        del obj.env_file

        config_test_yml = tmp_path / 'config-test.yml'
        config_test_yml.touch()
        assert obj.env_file == config_test_yml
        del obj.env_file

        env_test_yml = env_dir / 'test.yml'
        env_test_yml.touch()
        assert obj.env_file == env_test_yml
        del obj.env_file

        config_test_us_east_1_yml = tmp_path / 'config-test-us-east-1.yml'
        config_test_us_east_1_yml.touch()
        assert obj.env_file == config_test_us_east_1_yml
        del obj.env_file

        env_test_us_east_1_yml = env_dir / 'test-us-east-1.yml'
        env_test_us_east_1_yml.touch()
        assert obj.env_file == env_test_us_east_1_yml
        del obj.env_file

    @patch('runway.module.serverless.merge_dicts')
    def test_extend_serverless_yml(self, mock_merge, caplog, monkeypatch,
                                   runway_context, tmp_path):
        """Test extend_serverless_yml."""
        # pylint: disable=no-member
        caplog.set_level(logging.DEBUG, logger='runway')
        mock_func = MagicMock()
        mock_merge.return_value = {'key': 'val'}
        monkeypatch.setattr(Serverless, 'npm_install', MagicMock())
        monkeypatch.setattr(Serverless, 'sls_print',
                            MagicMock(return_value='original'))
        monkeypatch.setattr(ServerlessOptions, 'update_args', MagicMock())

        options = {
            'extend_serverless_yml': {'new-key': 'val'}
        }
        obj = Serverless(runway_context, tmp_path,
                         options={'options': options})

        assert not obj.extend_serverless_yml(mock_func)
        obj.npm_install.assert_called_once()
        obj.sls_print.assert_called_once()
        mock_merge.assert_called_once_with('original',
                                           options['extend_serverless_yml'])
        mock_func.assert_called_once_with(skip_install=True)
        obj.options.update_args.assert_called_once_with('config', ANY)

        tmp_file = obj.options.update_args.call_args[0][1]
        # 'no way to check the prefix since it will be a uuid'
        assert tmp_file.endswith('.tmp.serverless.yml')
        assert not (tmp_path / tmp_file).exists(), \
            'should always be deleted after calling "func"'

        caplog.clear()
        monkeypatch.setattr('{}.Path.unlink'.format(
            'pathlib' if sys.version_info.major == 3 else 'pathlib2'
        ), MagicMock(side_effect=OSError('test OSError')))
        assert not obj.extend_serverless_yml(mock_func)
        assert '{}: test OSError'.format(tmp_path.name) in caplog.messages

    @patch('runway.module.serverless.generate_node_command')
    @pytest.mark.parametrize('command', [('deploy'), ('remove'), ('print')])
    def test_gen_cmd(self, mock_cmd, command, monkeypatch, runway_context,
                     tmp_path):
        """Test gen_cmd."""
        # pylint: disable=no-member
        monkeypatch.setattr(Serverless, 'log_npm_command', MagicMock())
        monkeypatch.setattr(runway_context, 'no_color', False)
        mock_cmd.return_value = ['success']
        obj = Serverless(runway_context, tmp_path,
                         {'options': {'args': ['--config', 'test']}})
        expected_opts = [command,
                         '--region', runway_context.env_region,
                         '--stage', runway_context.env_name,
                         '--config', 'test',
                         '--extra-arg']

        assert obj.gen_cmd(command, args_list=['--extra-arg']) == ['success']
        mock_cmd.assert_called_once_with(command='sls',
                                         command_opts=expected_opts,
                                         path=tmp_path)
        obj.log_npm_command.assert_called_once_with(['success'])
        mock_cmd.reset_mock()

        obj.context.env_vars['CI'] = '1'
        monkeypatch.setattr(runway_context, 'no_color', True)
        expected_opts.append('--no-color')
        if command not in ['remove', 'print']:
            expected_opts.append('--conceal')
        assert obj.gen_cmd(command, args_list=['--extra-arg']) == ['success']
        mock_cmd.assert_called_once_with(command='sls',
                                         command_opts=expected_opts,
                                         path=tmp_path)

    def test_init(self, caplog, runway_context):
        """Test init and the attributes set in init."""
        caplog.set_level(logging.ERROR, logger='runway')
        obj = Serverless(runway_context, './tests',
                         {'options': {'skip_npm_ci': True}})
        assert isinstance(obj.options, ServerlessOptions)
        assert obj.region == runway_context.env_region
        assert obj.stage == runway_context.env_name

        with pytest.raises(SystemExit):
            assert not Serverless(runway_context, './tests',
                                  {'options': {
                                      'promotezip': {'invalid': 'value'}
                                  }})
        assert ['tests: "bucketname" must be provided when using '
                '"options.promotezip": {\'invalid\': \'value\'}'] == caplog.messages

    def test_plan(self, caplog, runway_context):
        """Test plan."""
        caplog.set_level(logging.INFO, logger='runway')
        obj = Serverless(runway_context, './tests')

        assert not obj.plan()
        assert ['Planning not currently supported for Serverless'] == \
            caplog.messages

    def test_skip(self, caplog, monkeypatch, runway_context, tmp_path):
        """Test skip."""
        caplog.set_level(logging.WARNING, logger='runway')
        obj = Serverless(runway_context, tmp_path)
        monkeypatch.setattr(obj, 'package_json_missing', lambda: True)
        monkeypatch.setattr(obj, 'env_file', False)

        assert obj.skip
        assert ['{}: The Serverless module type requires a package file '
                'specifying serverless in devDependencies'.format(tmp_path.name),
                '{}: Skipping module'.format(tmp_path.name)] == caplog.messages
        caplog.clear()

        monkeypatch.setattr(obj, 'package_json_missing', lambda: False)
        assert obj.skip
        assert ['{}: No config file for this stage/region found (looking for '
                'one of "{}")'.format(tmp_path.name,
                                      ', '.join(gen_sls_config_files(obj.stage,
                                                                     obj.region))),
                '{}: Skipping module'.format(tmp_path.name)] == caplog.messages
        caplog.clear()

        obj.environments = True
        assert not obj.skip
        obj.environments = False

        obj.parameters = True
        assert not obj.skip
        obj.parameters = False

        obj.env_file = True
        assert not obj.skip

    @patch('runway.module.serverless.deploy_package')
    @patch('runway.module.serverless.run_module_command')
    def test_sls_deploy(self, mock_run, mock_deploy, monkeypatch,
                        runway_context, tmp_path):
        """Test sls_deploy."""
        # pylint: disable=no-member
        monkeypatch.setattr(runway_context, 'no_color', False)
        monkeypatch.setattr(Serverless, 'gen_cmd',
                            MagicMock(return_value=['deploy']))
        monkeypatch.setattr(Serverless, 'npm_install', MagicMock())
        obj = Serverless(runway_context, tmp_path,
                         options={'options': {'args': ['--config', 'test.yml']}})

        assert not obj.sls_deploy()
        obj.npm_install.assert_called_once()
        obj.gen_cmd.assert_called_once_with('deploy')
        mock_run.assert_called_once_with(cmd_list=['deploy'],
                                         env_vars=runway_context.env_vars)

        obj.options.promotezip['bucketname'] = 'test-bucket'
        assert not obj.sls_deploy(skip_install=True)
        obj.npm_install.assert_called_once()
        mock_deploy.assert_called_once_with(
            ['deploy',
             '--region', runway_context.env_region,
             '--stage', runway_context.env_name,
             '--config', 'test.yml'],
            'test-bucket',
            runway_context,
            str(tmp_path)
        )
        mock_run.assert_called_once()

        monkeypatch.setattr(runway_context, 'no_color', True)
        assert not obj.sls_deploy(skip_install=True)
        mock_deploy.assert_called_with(
            ['deploy',
             '--region', runway_context.env_region,
             '--stage', runway_context.env_name,
             '--config', 'test.yml',
             '--no-color'],
            'test-bucket',
            runway_context,
            str(tmp_path)
        )

    def test_sls_print(self, monkeypatch, runway_context):
        """Test sls_print."""
        # pylint: disable=no-member
        expected_dict = {'status': 'success'}
        mock_check_output = MagicMock(return_value=yaml.safe_dump(expected_dict))
        monkeypatch.setattr(Serverless, 'gen_cmd',
                            MagicMock(return_value=['print']))
        monkeypatch.setattr(Serverless, 'npm_install', MagicMock())
        monkeypatch.setattr('subprocess.check_output', mock_check_output)
        obj = Serverless(runway_context, './tests')

        assert obj.sls_print() == expected_dict
        obj.npm_install.assert_called_once()
        mock_check_output.assert_called_once_with(['print'],
                                                  env=runway_context.env_vars)
        obj.gen_cmd.assert_called_once_with('print',
                                            args_list=['--format', 'yaml'])
        obj.gen_cmd.reset_mock()

        assert obj.sls_print(item_path='something.status',
                             skip_install=True) == expected_dict
        obj.npm_install.assert_called_once()
        obj.gen_cmd.assert_called_once_with('print',
                                            args_list=['--format', 'yaml',
                                                       '--path', 'something.status'])

    def test_sls_remove(self, monkeypatch, runway_context):
        """Test sls_remove."""
        # pylint: disable=no-member
        # TODO use pytest-subprocess for when dropping python 2
        sls_error = [
            '  Serverless Error ---------------------------------------',
            '',
            "  Stack 'test-stack' does not exist",
            '',
            '  Get Support --------------------------------------------',
            '     Docs:          docs.serverless.com',
            '     Bugs:          github.com/serverless/serverless/issues',
            '     Issues:        forum.serverless.com'
        ]
        mock_popen = MagicMock(return_value=MockProcess(returncode=0,
                                                        stdout='success'))
        monkeypatch.setattr('subprocess.Popen', mock_popen)
        monkeypatch.setattr(Serverless, 'gen_cmd',
                            MagicMock(return_value=['remove']))
        monkeypatch.setattr(Serverless, 'npm_install', MagicMock())

        obj = Serverless(runway_context, './tests')
        assert not obj.sls_remove()
        obj.npm_install.assert_called_once()
        obj.gen_cmd.assert_called_once_with('remove')
        mock_popen.assert_called_once_with(['remove'],
                                           bufsize=1,
                                           env=runway_context.env_vars,
                                           stdout=subprocess.PIPE,
                                           universal_newlines=True)
        mock_popen.return_value.wait.assert_called_once()

        mock_popen = MagicMock(return_value=MockProcess(returncode=1,
                                                        stdout=sls_error))

        monkeypatch.setattr('subprocess.Popen', mock_popen)
        assert not obj.sls_remove(skip_install=True)
        obj.npm_install.assert_called_once()
        mock_popen.return_value.wait.assert_called_once()

        sls_error[2] = '  Some other error'
        mock_popen = MagicMock(return_value=MockProcess(returncode=1,
                                                        stdout=sls_error))
        monkeypatch.setattr('subprocess.Popen', mock_popen)
        with pytest.raises(SystemExit):
            assert not obj.sls_remove()
        mock_popen.return_value.wait.assert_called_once()


class TestServerlessOptions(object):
    """Test runway.module.serverless.ServerlessOptions."""

    @pytest.mark.parametrize('args, expected', [
        (['--config', 'something'], ['--config', 'something']),
        (['--config', 'something', '--unknown-arg', 'value'],
         ['--config', 'something', '--unknown-arg', 'value']),
        (['-c', 'something'], ['--config', 'something']),
        (['-u'], ['-u'])
    ])
    def test_args(self, args, expected):
        """Test args."""
        obj = ServerlessOptions(args=args,
                                extend_serverless_yml={},
                                promotezip={})
        assert obj.args == expected

    @pytest.mark.parametrize('config', [
        ({'args': ['--config', 'something']}),
        ({'extend_serverless_yml': {'new_key': 'test_value'}}),
        ({'promotezip': {'bucketname': 'test-bucket'}}),
        ({'skip_npm_ci': True}),
        ({'args': ['--config', 'something'],
          'extend_serverless_yml': {'new_key': 'test_value'}}),
        ({'args': ['--config', 'something'],
          'extend_serverless_yml': {'new_key': 'test_value'},
          'promotezip': {'bucketname': 'test-bucket'}}),
        ({'args': ['--config', 'something'],
          'extend_serverless_yml': {'new_key': 'test_value'},
          'promotezip': {'bucketname': 'test-bucket'},
          'skip_npm_ci': True}),
        ({'args': ['--config', 'something'],
          'extend_serverless_yml': {'new_key': 'test_value'},
          'promotezip': {'bucketname': 'test-bucket'},
          'skip_npm_ci': False})
    ])
    def test_parse(self, config):
        """Test parse."""
        obj = ServerlessOptions.parse(**config)

        assert obj.args == config.get('args', [])
        assert obj.extend_serverless_yml == \
            config.get('extend_serverless_yml', {})
        assert obj.promotezip == config.get('promotezip', {})
        assert obj.skip_npm_ci == config.get('skip_npm_ci', False)

    def test_parse_invalid_promotezip(self):
        """Test parse with invalid promotezip value."""
        with pytest.raises(ValueError):
            assert not ServerlessOptions.parse(promotezip={'key': 'value'})

    def test_update_args(self):
        """Test update_args."""
        obj = ServerlessOptions(args=['--config', 'something',
                                      '--unknown-arg', 'value'],
                                extend_serverless_yml={},
                                promotezip={})
        assert obj.args == ['--config', 'something',
                            '--unknown-arg', 'value']

        obj.update_args('config', 'something-else')
        assert obj.args == ['--config', 'something-else',
                            '--unknown-arg', 'value']

        with pytest.raises(KeyError):
            obj.update_args('invalid-key', 'anything')



#!/usr/bin/env python
"""Module with IAM role."""
from __future__ import print_function

from troposphere import Join, iam

import awacs.sts

from awacs.aws import Allow, PolicyDocument, Principal, Statement

from stacker.blueprints.base import Blueprint
from stacker.blueprints.variables.types import CFNString

TESTING_ACCOUNT_ID = '523485371024'


class CrossAccountRole(Blueprint):
    """Stacker blueprint for IAM role."""

    VARIABLES = {
        'EnvironmentName': {'type': CFNString,
                            'description': 'Name of environment'}
    }

    def create_template(self):
        """Create template (main function called by Stacker)."""
        template = self.template
        variables = self.get_variables()
        template.set_version('2010-09-09')
        template.set_description('Runway Integration Testing - IAM Role')

        # Resources
        template.add_resource(
            iam.Role(
                'CodeBuildRole',
                AssumeRolePolicyDocument=PolicyDocument(
                    Statement=[
                        Statement(
                            Effect=Allow,
                            Action=[awacs.sts.AssumeRole],
                            Principal=Principal(
                                'AWS',
                                TESTING_ACCOUNT_ID
                            )
                        )
                    ]
                ),
                Description='Role used for cross account testing in runway',
                ManagedPolicyArns=[
                    'arn:aws:iam::aws:policy/AdministratorAccess'
                ],
                RoleName=Join('-', ['runway-integration-test-role',
                                    variables['EnvironmentName'].ref])
            )
        )


if __name__ == "__main__":
    from stacker.context import Context
    print(CrossAccountRole('test', Context({"namespace": "test"}), None).to_json())



"""Integration test module."""
import errno
import os
import shutil
import subprocess
import sys
from copy import deepcopy

import yaml
from send2trash import send2trash

from runway.util import change_dir


class IntegrationTest(object):
    """Base class for Integration Tests.

    Attributes:
        REQUIRED_FIXTURE_FILES (List[str]): List of fixture files that
            will be copied to the current ``working_dir`` from
            ``fixture_dir`` when using the ``copy_fixtures`` method.
        fixture_dir (str): Path to ``fixture`` directory relative to
            ``working_dir``.
        tests_dir (str): Path to ``tests`` directory relative to
            ``working_dir``.
        working_dir (str): Path that the test is running in.

    """

    REQUIRED_FIXTURE_FILES = []

    def __init__(self, logger, env_vars=None):
        """Initialize base class."""
        self.logger = logger
        self.environment = deepcopy(env_vars or os.environ)
        self.runway_config_path = None
        # roundabout way to get the file path of a subclass
        self.working_dir = os.path.abspath(os.path.dirname(
            sys.modules[self.__module__].__file__
        ))
        self.fixture_dir = os.path.join(self.working_dir, 'fixtures')
        self.tests_dir = os.path.join(self.working_dir, 'tests')

    @property
    def deploy_env(self):
        """Return DEPLOY_ENVIRONMENT for the test."""
        return self.environment.get('DEPLOY_ENVIRONMENT')

    @deploy_env.setter
    def deploy_env(self, value):
        """Set the DEPLOY_ENVIRONMENT value."""
        self.logger.info('Setting "DEPLOY_ENVIRONMENT" to "%s"', value)
        self.environment['DEPLOY_ENVIRONMENT'] = value

    @property
    def region(self):
        """Return the region set for the test environment."""
        return self.environment.get('AWS_DEFAULT_REGION',
                                    self.environment.get('AWS_REGION',
                                                         'us-east-1'))

    @region.setter
    def region(self, value):
        """Set the value of region."""
        self.environment['AWS_DEFAULT_REGION'] = value
        self.environment['AWS_REGION'] = value

    def copy_fixtures(self):
        """Copy fixtures to the root of the tests dir."""
        self.logger.info('Fixtures defined for tests: %s',
                         str(self.REQUIRED_FIXTURE_FILES))
        for fixture in self.REQUIRED_FIXTURE_FILES:
            src = os.path.join(self.fixture_dir, fixture)
            dest = os.path.join(self.working_dir, fixture)
            if os.path.isfile(src):
                self.logger.info('Copying "%s" to "%s"...', src, dest)
                shutil.copy(src, dest)

    def cleanup_fixtures(self):
        """Delete copied fixtures."""
        for fixture in self.REQUIRED_FIXTURE_FILES:
            fixture_path = os.path.join(self.working_dir, fixture)
            self.logger.info('Deleting "%s"...', fixture_path)
            try:
                send2trash(fixture_path)
            except OSError as err:
                if err.errno == errno.ENOENT or 'not found' in str(err):
                    continue
                raise

    def parse_config(self, path):
        """Read and parse yml."""
        if not os.path.isfile(path):
            self.logger.error("Config file was not found (looking for \"%s\")",
                              path)
        with open(path) as data_file:
            return yaml.safe_load(data_file)

    def runway_cmd(self, action, *args, env_vars=None, tags=None, timeout=900):
        """Run a deploy command based on tags.

        Args:
            action (str): Runway action. (e.g. ``deploy``, ``destroy``)
            env_vars (Optional[Dict[str, str]]): Can be used to override
                environment variables for the invocation.
            tags (Optional[List[str]]): List of tag options to pass to Runway.
            timeout (int): Seconds to wait for process to complete.
            args (str): Additional arguments to add to the command. These
                are places after any ``--tag`` options.

        Returns:
            Tuple[int, str, str]: The return code, ``stdout``, and ``stderr``
            of the process.

        """
        cmd = ['runway', action]
        if tags:
            for tag in tags:
                cmd.extend(['--tag', tag])
        cmd.extend(args)
        self.logger.info('Running command: %s', str(cmd))
        with change_dir(self.working_dir):
            cmd_process = subprocess.Popen(cmd, env=env_vars or self.environment,
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE,
                                           universal_newlines=True)
            stdout, stderr = cmd_process.communicate(timeout=timeout)
            print(stderr)
        return cmd_process.returncode, stdout, stderr

    def set_environment(self, env):
        """Set deploy environment."""
        if isinstance(env, str):
            self.deploy_env = env

    def set_env_var(self, var_name, var):
        """Set an environment variable"""
        self.logger.info('Setting "%s" to "%s"', var_name, var)
        if not isinstance(var, dict):
            env = {}
            env[var_name] = var
        self.environment.update(env)

    def unset_env_var(self, var):
        self.logger.info('Unsetting "%s" Environment Variable', var)
        del self.environment[var]

    def run(self):
        """Implement dummy method (set in consuming classes)."""
        raise NotImplementedError('You must implement the run() method '
                                  'yourself!')

    def teardown(self):
        """Implement dummy method (set in consuming classes)."""
        raise NotImplementedError('You must implement the teardown() method '
                                  'yourself!')



"""Test Runway config classes."""
# pylint: disable=no-self-use
import logging
import os
from copy import deepcopy
from tempfile import NamedTemporaryFile

import pytest
import yaml
from runway.cfngin.exceptions import UnresolvedVariable
from runway.config import (DeploymentDefinition, FutureDefinition,
                           ModuleDefinition)
# tries to test the imported class unless using "as"
from runway.config import TestDefinition as ConfigTestDefinition
from runway.config import VariablesDefinition
from runway.util import MutableMap

YAML_FIXTURES = ['config.runway.yml', 'config.runway.variables.yml']
ENV_VARS = {
    'AWS_REGION': 'us-east-1',
    'DEPLOY_ENVIRONMENT': 'test',
    'USER': 'test'
}


class TestDeploymentDefinition(object):
    """Test DeploymentDefinition."""

    ATTRS = ['_account_alias', '_account_id', '_assume_role',
             '_env_vars', '_environments', '_module_options', 'modules',
             'name', '_regions', '_parallel_regions']

    def test_from_list(self, yaml_fixtures):
        """Test init of a deployment from a list."""
        raw_config = deepcopy(yaml_fixtures['config.runway.yml']['deployments'])
        deployment = DeploymentDefinition.from_list(raw_config)[0]
        deployment_attrs = deployment.__dict__.keys()

        assert deployment.name == 'deployment_1'

        for attr in self.ATTRS:  # provides a better error than using all()
            assert attr in deployment_attrs

    def test_pre_process_resolve(self, yaml_fixtures):
        """Test that pre-process resolution only resolves specific vars."""
        raw_config = deepcopy(yaml_fixtures['config.runway.yml']['deployments'])
        raw_vars = deepcopy(yaml_fixtures['config.runway.variables.yml'])
        deployment = DeploymentDefinition.from_list(raw_config)[0]
        raw_context = {'env_vars': os.environ.copy()}
        raw_context['env_vars'].update(ENV_VARS)
        deployment.resolve(MutableMap(**raw_context),
                           variables=MutableMap(**raw_vars),
                           pre_process=True)

        # check resolved variables for pre_process
        assert not deployment.account_id == 123456789101
        assert deployment.assume_role['arn'] == 'arn:aws:iam::role/some-role'
        assert deployment.env_vars == {'MY_USERNAME': 'test'}
        assert deployment.regions == ['us-east-1']

        assert not deployment.parallel_regions, 'not set in test config, should be None'

        # these should be unresolved at this point
        for attr in ['environments', 'module_options']:
            with pytest.raises(UnresolvedVariable):
                getattr(deployment, attr)

    def test_resolve(self, yaml_fixtures):
        """Test full resolution of variable attributes."""
        raw_config = deepcopy(yaml_fixtures['config.runway.yml']['deployments'])
        raw_vars = deepcopy(yaml_fixtures['config.runway.variables.yml'])
        deployment = DeploymentDefinition.from_list(raw_config)[0]
        raw_context = {'env_vars': os.environ.copy()}
        raw_context['env_vars'].update(ENV_VARS)
        deployment.resolve(MutableMap(**raw_context),
                           variables=MutableMap(**raw_vars))

        assert deployment.regions == ['us-east-1']
        assert not deployment.account_id == 123456789101
        assert deployment.assume_role['arn'] == 'arn:aws:iam::role/some-role'
        assert deployment.env_vars == {'MY_USERNAME': 'test'}
        assert deployment.environments == {
            'test_param': 'lab value for ${envvar AWS_REGION}'
        }
        assert deployment.module_options == {
            'deployment_option': 'test.deployment.module_options'
        }
        assert deployment.regions == ['us-east-1']

        assert not deployment.parallel_regions, 'not set in test config, should be None'


class TestFutureDefinition(object):
    """Test FutureDefinition."""

    def test_init(self, caplog):
        """Test init and the attributes it sets."""
        caplog.set_level(logging.INFO, logger='runway')
        config = {
            'strict_environments': True,
            'invalid_key': True
        }

        result = FutureDefinition(**config)
        assert result.strict_environments is config['strict_environments']
        assert caplog.messages == [
            'Invalid key(s) found in "future" have been ignored: invalid_key'
        ]

        with pytest.raises(TypeError):
            assert not FutureDefinition(strict_environments='true')
        assert not any(val for val in FutureDefinition().data.values())

    @pytest.mark.parametrize('config, expected', [
        ({'strict_environments': True}, ['strict_environments']),
        ({'strict_environments': False}, [])
    ])
    def test_enabled(self, config, expected):
        """Tested enabled."""
        assert FutureDefinition(**config).enabled == expected


class TestModuleDefinition(object):
    """Test ModuleDefinition."""

    ATTRS = ['child_modules', '_class_path', '_env_vars', '_environments',
             'name', '_options', '_path', 'tags']

    def test_from_list(self, yaml_fixtures):
        """Test init of a module from a list."""
        raw_config = deepcopy(
            yaml_fixtures['config.runway.yml']['deployments'][0]['modules'][0]
        )
        module = ModuleDefinition.from_list(raw_config)[0]
        module_attrs = module.__dict__.keys()

        for attr in self.ATTRS:  # provides a better error than using all()
            assert attr in module_attrs

    def test_resolve(self, yaml_fixtures):
        """Test full resolution of variable attributes."""
        raw_config = deepcopy(
            yaml_fixtures['config.runway.yml']['deployments'][0]['modules']
        )
        raw_vars = deepcopy(yaml_fixtures['config.runway.variables.yml'])
        module = ModuleDefinition.from_list(raw_config)[0]
        raw_context = {'env_vars': os.environ.copy()}
        raw_context['env_vars'].update(ENV_VARS)
        module.resolve(MutableMap(**raw_context),
                       variables=MutableMap(**raw_vars))

        assert module.child_modules == []
        assert not module.class_path
        assert module.env_vars == {'MY_USERNAME_MODULE': 'test'}
        assert module.environments == {'module_test_param': 'test'}
        assert module.name == '${var test_path}app.cfn'
        assert module.options == {'sample_module_option': 'test.module.options'}
        assert module.path == 'sampleapp.cfn'
        assert module.tags == {}


class TestTestDefinition(object):
    """Test TestDefinition."""

    ATTRS = ['_args', 'name', '_required', 'type']

    def test_from_list(self, yaml_fixtures):
        """Test init of a deployment from a list."""
        raw_config = deepcopy(yaml_fixtures['config.runway.yml']['tests'])
        test = ConfigTestDefinition.from_list(raw_config)[0]
        test_attrs = test.__dict__.keys()

        for attr in self.ATTRS:  # provides a better error than using all()
            assert attr in test_attrs

    def test_resolve(self, yaml_fixtures):
        """Test full resolution of variable attributes."""
        raw_config = deepcopy(yaml_fixtures['config.runway.yml']['tests'])
        raw_vars = deepcopy(yaml_fixtures['config.runway.variables.yml'])
        test = ConfigTestDefinition.from_list(raw_config)[0]
        raw_context = {'env_vars': os.environ.copy()}
        raw_context['env_vars'].update(ENV_VARS)
        test.resolve(MutableMap(**raw_context),
                     variables=MutableMap(**raw_vars))

        assert test.args == {'commands': ['echo "My name is test"']}
        assert test.name == 'hello_world'
        assert isinstance(test.required, bool)
        assert test.required
        assert test.type == 'script'


class TestVariablesDefinition(object):
    """Test VariablesDefinition."""

    def test_load(self, yaml_fixtures):
        """Test loading variables from a file with given path.

        Also tests setting variable from kwargs.

        """
        with NamedTemporaryFile(mode='w+', suffix='.yml') as var_file:
            var_file.write(
                yaml.safe_dump(yaml_fixtures['config.runway.variables.yml'])
            )
            var_file.seek(0)  # return curser to the top of the file
            result = VariablesDefinition.load(
                file_path=var_file.name, explicit_kwarg='not in file'
            )

            assert result.test_value == 'basic value'
            assert result.explicit_kwarg == 'not in file'

    def test_load_explicit_file_missing(self, caplog):
        """Test missing explicit file results in an error."""
        caplog.set_level('ERROR', logger='runway')

        with pytest.raises(SystemExit):
            VariablesDefinition.load(file_path='fake_file.yaml')

        assert caplog.records[0].msg == ('The provided variables "%s" file '
                                         'could not be found.')

    def test_load_no_file(self, caplog):
        """Should not error when default variables file is not found."""
        caplog.set_level('INFO', logger='runway')
        result = VariablesDefinition.load()

        assert result.data == {}
        assert caplog.records[0].msg == ('Could not find %s in the current '
                                         'directory. Continuing without a '
                                         'variables file.')



"""Tests for the Path type object."""
import logging
import unittest

from runway.sources.source import Source
from runway.path import Path

LOGGER = logging.getLogger('runway')


class MockGitSource(Source):
    """Mock a Git Source Object."""

    def fetch(self):
        """Fetch."""
        return 'mock/git/folder'


class PathTester(unittest.TestCase):
    """Test Path class."""

    def test_parse_local_source_string(self):
        """Parsing location source string. Verify tuple is parsed as anticipated."""
        (source, uri, location, options) = Path.parse({'path': 'src/foo/bar'})
        self.assertEqual(source, 'local')
        self.assertEqual(location, 'src/foo/bar')
        self.assertEqual(uri, '')
        self.assertEqual(options, {})

    def test_parse_git_source_no_location_or_options(self):
        """Parsing Git source with no location or options.

        Verify tuple is parsed as anticipated.

        """
        (source, uri, location, options) = Path.parse(
            {'path': 'git::git://github.com/onicagroup/foo/bar.git'}
        )
        self.assertEqual(source, 'git')
        self.assertEqual(location, '')
        self.assertEqual(uri, 'git://github.com/onicagroup/foo/bar.git')
        self.assertEqual(options, {})

    def test_parse_git_source_with_location_no_options(self):
        """Parsing Git source with location, no options. Verify tuple is parsed as anticipated."""
        (source, uri, location, options) = Path.parse(
            {'path': 'git::git://github.com/onicagroup/foo/bar.git//foo/bar'}
        )
        self.assertEqual(source, 'git')
        self.assertEqual(location, 'foo/bar')
        self.assertEqual(uri, 'git://github.com/onicagroup/foo/bar.git')
        self.assertEqual(options, {})

    def test_parse_git_source_with_options_no_location(self):
        """
        Parsing Git source with options, no location.

        Verify tuple is parsed as anticipated.
        """
        (source, uri, location, options) = Path.parse(
            {'path': 'git::git://github.com/onicagroup/foo/bar.git?branch=foo'}
        )
        self.assertEqual(source, 'git')
        self.assertEqual(location, '')
        self.assertEqual(uri, 'git://github.com/onicagroup/foo/bar.git')
        self.assertEqual(options, {'branch': 'foo'})

    def test_parse_git_source_with_multiple_options_no_location(self):
        """
        Parsing Git source with multiple options, no location.

        Verify tuple is parsed as anticipated.
        """
        (source, uri, location, options) = Path.parse(
            {'path': 'git::git://github.com/onicagroup/foo/bar.git?branch=foo&bar=baz'}
        )
        self.assertEqual(source, 'git')
        self.assertEqual(location, '')
        self.assertEqual(uri, 'git://github.com/onicagroup/foo/bar.git')
        self.assertEqual(options, {'branch': 'foo', 'bar': 'baz'})

    def test_parse_git_source_with_options_and_location(self):
        """Parsing Git source with options and location.

        Verify tuple is parsed as anticipated.

        """
        (source, uri, location, options) = Path.parse(
            {'path': 'git::git://github.com/onicagroup/foo/bar.git//src/foo/bar?branch=foo'}
        )
        self.assertEqual(source, 'git')
        self.assertEqual(location, 'src/foo/bar')
        self.assertEqual(uri, 'git://github.com/onicagroup/foo/bar.git')
        self.assertEqual(options, {'branch': 'foo'})

    def test_parse_git_source_with_multiple_options_and_location(self):
        """Parsing Git source with multiple options and location.

        Verify tuple is parsed as anticipated.

        """
        path = 'git::git://github.com/onicagroup/foo/bar.git//src/foo/bar?branch=foo&bar=baz'
        (source, uri, location, options) = Path.parse({'path': path})
        self.assertEqual(source, 'git')
        self.assertEqual(location, 'src/foo/bar')
        self.assertEqual(uri, 'git://github.com/onicagroup/foo/bar.git')
        self.assertEqual(options, {'branch': 'foo', 'bar': 'baz'})

    def test_configuration_property(self):
        """Verify the Configuration property is set to appropriate values."""
        path = 'git::git://github.com/onicagroup/foo/bar.git//src/foo/bar?branch=foo'
        instance = Path({'path': path}, 'fake/env/root', git_source_class=MockGitSource)
        self.assertEqual(instance.configuration, {
            'source': 'git',
            'location': 'src/foo/bar',
            'uri': 'git://github.com/onicagroup/foo/bar.git',
            'options': {'branch': 'foo'},
            'cache_dir': None
        })

    def test_module_root_set_to_env_root(self):
        """When the path location == a root directory set to the env_root passed."""
        path = '.'
        instance = Path({'path': path}, 'fake/env/root', git_source_class=MockGitSource)
        self.assertEqual(instance.module_root, 'fake/env/root')

        path = './'
        instance2 = Path({'path': path}, 'fake/env/root', git_source_class=MockGitSource)
        self.assertEqual(instance2.module_root, 'fake/env/root')

    def test_module_root_set_to_contcatenated_path(self):
        """Test module root set to concatednated path.

        When the path location is local and not the root combine the location
        with the env_root.

        """
        path = 'foo/bar'
        instance = Path({'path': path}, 'fake/env/root', git_source_class=MockGitSource)
        self.assertEqual(instance.module_root, 'fake/env/root/foo/bar')

    def test_module_root_set_to_fetched_source_value(self):
        """When the path location is a remote resource fetch the directory."""
        path = 'git::git://github.com/onicagroup/foo/bar.git//src/foo/bar?branch=foo'
        instance = Path({'path': path}, 'fake/env/root', git_source_class=MockGitSource)
        self.assertEqual(instance.module_root, 'mock/git/folder')



"""Output ``runway.yml``-defined environment variables.

OS environment variables can be set in ``runway.yml`` for different
Runway environments (e.g. dev & prod ``KUBECONFIG`` values). The
``envvars`` command allows access to these values for use outside of
Runway.

Example:
  .. code-block:: shell

    $ eval "$(runway envvars)"

"""
from __future__ import print_function

import logging
import os
import platform
import sys
from builtins import input  # python 2 compatability
from typing import Dict, List  # noqa pylint: disable=unused-import

from ...cfngin.exceptions import FailedVariableLookup
from ...context import Context
from ...util import merge_dicts, merge_nested_environment_dicts
from ..runway_command import RunwayCommand, get_env

LOGGER = logging.getLogger('runway')


def print_env_vars_psh(env_vars):
    # type: (Dict[str, str]) -> None
    """Print environment variables for Powershell."""
    for (key, val) in env_vars.items():
        print("$env:%s =\"%s\"" % (key, val))


def print_env_vars_posix(env_vars):
    # type: (Dict[str, str]) -> None
    """Print environment variables for bash."""
    for (key, val) in env_vars.items():
        print("export %s=\"%s\"" % (key, val))


def select_region(context, options):
    # type: (Context, List[str]) -> Context
    """Prompt the user to select an AWS region.

    The context object is updated with the selected region.

    Args:
        context (Context): Runway context object.
        options (List[str]): List of AWS regions to select from.

    Returns:
        Context: The provided context object updated with the selected region.

    """
    if len(options) == 1:
        result = options[0]
    else:
        print('', file=sys.stderr)
        print('Configured Regions:', file=sys.stderr)
        for i, opt in enumerate(options):
            print(' {}: {}'.format(i + 1, opt), file=sys.stderr)
        selection = None
        while not selection:
            try:
                selection = input('Enter number of region to use: ')
                # separate type change for more informative error
                selection = int(selection)
                result = options[selection - 1 if selection > 0 else selection]
            except (IndexError, ValueError):
                LOGGER.error('Value must be between 1 and %s; got "%s"',
                             len(options), selection)
                sys.exit(1)
        print('', file=sys.stderr)
    context.env_region = result
    context.env_vars.update({'AWS_DEFAULT_REGION': result,
                             'AWS_REGION': result})
    return context


class EnvVars(RunwayCommand):
    """Extend RunwayCommand with execution of envvars."""

    def execute(self):
        """Output Runway-defined environment variables."""
        if self.runway_config.deployments:
            context = Context(
                env_name=get_env(self.env_root,
                                 self.runway_config.ignore_git_branch),
                env_region=None,
                env_root=self.env_root,
                env_vars=os.environ.copy()
            )
            env_vars = {}
            for deployment in self.runway_config.deployments:
                try:
                    deployment.resolve(context, self.runway_vars)
                except FailedVariableLookup as err:
                    if any(var in str(err.lookup)
                           for var in ['AWS_REGION', 'AWS_DEFAULT_REGION']):
                        select_region(context, deployment.regions)
                        deployment.resolve(context, self.runway_vars)
                    else:
                        raise
                env_vars = merge_dicts(env_vars, deployment.env_vars)
            if env_vars:
                env_vars = merge_nested_environment_dicts(
                    env_vars,
                    env_name=context.env_name,
                    env_root=context.env_root
                )

            if 'MSYSTEM' in os.environ and (
                    os.environ['MSYSTEM'].startswith('MINGW') and (
                        platform.system() == 'Windows')):  # type: ignore
                print_env_vars_posix(env_vars)  # git bash
            elif platform.system() == 'Windows':  # type: ignore
                print_env_vars_psh(env_vars)
            else:
                print_env_vars_posix(env_vars)
        else:
            print('ERROR: no runway deployments found', file=sys.stderr)
            sys.exit(1)



"""Test runway.commands.runway_command."""
# pylint: disable=no-self-use
import os

import pytest
from git import InvalidGitRepositoryError
from mock import MagicMock, call, patch
from runway import __version__
from runway.commands.runway_command import (RunwayCommand, get_env,
                                            get_env_from_branch,
                                            get_env_from_directory,
                                            get_env_from_user)
from runway.util import environ

MODULE = 'runway.commands.runway_command'
PATCH_RUNWAY_CONFIG = 'runway.commands.base_command.Config'


@pytest.mark.usefixtures('patch_runway_config')
class TestRunwayCommand(object):
    """Test runway.commands.runway_command.RunwayCommand."""

    def test_execute(self):
        """Test execute."""
        with pytest.raises(NotImplementedError) as excinfo:
            assert RunwayCommand().execute()
        assert str(excinfo.value) == \
            'execute must be implimented for subclasses of BaseCommand.'

    def test_get_empty_dirs(self, tmp_path):
        """Test get_empty_dirs."""
        assert RunwayCommand.get_empty_dirs(str(tmp_path)) == []

        dir_1 = tmp_path / 'dir_1'
        dir_1.mkdir()
        dir_2 = tmp_path / 'dir_2'
        dir_2.mkdir()

        assert RunwayCommand.get_empty_dirs(str(tmp_path)) == ['dir_2',
                                                               'dir_1']

        (tmp_path / 'test_file').touch()
        assert RunwayCommand.get_empty_dirs(str(tmp_path)) == ['dir_2',
                                                               'dir_1']

        (tmp_path / '.git').mkdir()  # skip even if it is empty
        assert RunwayCommand.get_empty_dirs(str(tmp_path)) == ['dir_2',
                                                               'dir_1']

        (dir_2 / 'test_file').touch()
        assert RunwayCommand.get_empty_dirs(str(tmp_path)) == ['dir_1']

        (dir_1 / 'test_file').touch()
        assert RunwayCommand.get_empty_dirs(str(tmp_path)) == []

    def test_path_only_contains_dirs(self, tmp_path):
        """Test path_only_contains_dirs."""
        assert RunwayCommand.path_only_contains_dirs(str(tmp_path)) is True

        (tmp_path / 'test_file').touch()
        assert RunwayCommand.path_only_contains_dirs(str(tmp_path)) is False
        (tmp_path / 'test_file').unlink()

        dir_1 = tmp_path / 'dir_1'
        dir_1.mkdir()
        dir_2 = tmp_path / 'dir_2'
        dir_2.mkdir()
        assert RunwayCommand.path_only_contains_dirs(str(tmp_path)) is True

        (dir_1 / 'test_file').touch()
        assert RunwayCommand.path_only_contains_dirs(str(tmp_path)) is False

    def test_version(self, capsys):
        """Test version."""
        assert not RunwayCommand.version()
        assert capsys.readouterr().out == (__version__ + '\n')


@patch('git.Repo')
@patch('runway.commands.runway_command.get_env_from_branch')
@patch('runway.commands.runway_command.get_env_from_directory')
def test_get_env(mock_env_dir, mock_env_branch, mock_git_repo):
    """Test runway.commands.runway_command.get_env."""
    mock_repo = MagicMock()
    mock_repo.active_branch.name = 'git_branch'
    mock_git_repo.return_value = mock_repo
    mock_env_dir.return_value = 'dir_value'
    mock_env_branch.return_value = 'branch_value'

    with environ({}):
        os.environ.pop('DEPLOY_ENVIRONMENT', None)  # ensure not in env
        assert get_env('/', prompt_if_unexpected=True) == 'branch_value'
        mock_env_branch.assert_called_once_with('git_branch', True)

        # cases resulting in get_env_from_directory
        assert get_env('path', ignore_git_branch=True) == 'dir_value'
        mock_git_repo.side_effect = InvalidGitRepositoryError
        assert get_env('path') == 'dir_value'
        mock_env_dir.assert_has_calls([call('path'),
                                       call('path')])

        # handle TypeError
        mock_git_repo.side_effect = TypeError
        with pytest.raises(SystemExit):
            assert not get_env('path')
        mock_git_repo.side_effect = None

    with environ({'DEPLOY_ENVIRONMENT': 'test'}):
        assert get_env('path') == 'test'


@pytest.mark.parametrize('branch, do_prompt, did_prompt, expected', [
    ('ENV-dev', True, False, 'dev'),
    ('master', True, False, 'common'),
    ('invalid', True, True, 'user_value'),
    ('invalid', False, False, 'invalid')
])
def test_get_env_from_branch(branch, do_prompt, did_prompt, expected, monkeypatch):
    """Test runway.commands.runway_command.get_env_from_branch."""
    mock_prompt = MagicMock(return_value='user_value')
    monkeypatch.setattr('runway.commands.runway_command.get_env_from_user',
                        mock_prompt)
    assert get_env_from_branch(branch, do_prompt) == expected

    if did_prompt:
        mock_prompt.assert_called_once_with(branch)
    else:
        mock_prompt.assert_not_called()


def test_get_env_from_directory():
    """Test runway.commands.runway_command.get_env_from_directory."""
    assert get_env_from_directory('ENV-dev') == 'dev'
    assert get_env_from_directory('expected') == 'expected'


def test_get_env_from_user():
    """Test runway.commands.runway_command.get_env_from_user."""
    input_path = 'runway.commands.runway_command.input'

    with patch(input_path, MagicMock(return_value='n')):
        assert get_env_from_user('expected') == 'expected'

    with patch(input_path, MagicMock(side_effect=['y', '', 'expected'])):
        assert get_env_from_user('test') == 'expected'



from aws_cdk import (
    aws_iam as iam,
    aws_sqs as sqs,
    aws_sns as sns,
    aws_sns_subscriptions as subs,
    core
)

class MyStack1(core.Stack):

    def __init__(self, scope: core.Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        topic = sns.Topic(
            self, "MyFirstTopic",
            display_name="My First Topic"
        )



"""Test to verify behavior of explicit type declarations."""
import subprocess

from runway.util import change_dir

from integration_tests.test_runway_module_type_detection.test_runway_module_type_detection import (
    RunwayModuleTypeDetection
)


class TestTypeCloudformation(RunwayModuleTypeDetection):
    """Test to verify a 'type' definition of 'cloudformation' is respected."""

    TEST_NAME = __name__

    def deploy(self):
        """Deploy provider."""
        self.copy_fixture('type_cloudformation')
        self.copy_runway('type-cloudformation')
        with change_dir(self.mtd_test_dir):
            out = subprocess.check_output(['runway', 'deploy'], stderr=subprocess.STDOUT)
            return 0 if "Skipping stacker build" in out.decode() else -1

    def run(self):
        """Run tests."""
        self.clean()
        assert self.deploy() == 0, '{}: Type Declaration of Cloudformation Failed'.format(
            __name__
        )

    def teardown(self):
        """Teardown."""
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        self.clean()
